if not IsBound(_loaded) then _loaded:= false; fi; 

if not _loaded then
   LoadPackage("corelg");
   Reread("sqrt.gd");
   Reread("sqrt.gi");
   Reread("h1.gd");
   Reread("h1.gi");
   Read("levi.g");
   _loaded:= true;
fi;
   

printstuff:= true;

# the next function is from the library of corelg,
# we improve it by not requiring a sigma-fixed basis of H, just H needs to
# be fixed by sigma...

# first we have a function that finds a sigma-fixed basis of L, then
# in there a Cartan subalgebra

sigmaCSA:= function( L )

        local sigma, i, b0, sp, x, u, T, K, H;

        sigma := RealStructure( L );
        i:= E(4)*One( LeftActingDomain(L) );
        b0:= [ ];
        sp:= Subspace( L, [] );
        for x in Basis(L) do
            if sigma(x) = x then
               if not x in sp then
                  Add( b0, x );
                  sp:= Subspace( L, b0 );
               fi;
            else
               u:= x+sigma(x);
               if not u in sp then
                  Add( b0, u );
                  sp:= Subspace( L, b0 );
               fi;
               u:= i*(x-sigma(x));
               if not u in sp then
                  Add( b0, u );
                  sp:= Subspace( L, b0 );
               fi;
            fi;
        od;
        T:= StructureConstantsTable( Basis( L,b0 ) );
        K:= LieAlgebraByStructureConstants( LeftActingDomain(L), T );
        H:= CartanSubalgebra( K );
        return Subalgebra( L, List( Basis(H), x -> Coefficients( Basis(K),x )*
                      b0 ) );
end;

maxcptCSA:= function ( L )
    local F, sigma, H, R, pr, cb, cg, cbH, testIt, decomposeCSA, realRoot, 
    decH, alpha, i, x, y, j, K, prv, nrv, ev, rts, rt, rrr, rrv, spc, pos, 
    cf, sx;
    Info( InfoCorelg, 1, "start MaximallyCompactCartanSubalgebra" );
    F := LeftActingDomain( L );
    if not E( 4 ) * One( F ) in F then
        Error( "need E(4) in field" );
    fi;
    sigma := RealStructure( L );
    H := CartanSubalgebra( L );
    if not ForAll( Basis( H ), function ( x )
                 return sigma( x ) in H;
             end ) then

       if printstuff then
          Print("maxcptCSA: CSA not fixed by sigma, finding a new one...\n");
       fi;
       H:= sigmaCSA( L );
       if printstuff then
          Print("maxcptCSA: now CSA fixed by sigma: ",
             ForAll( Basis(H), x -> sigma(x) in H ), "\n");
       fi;
    fi;

    

    R := RootsystemOfCartanSubalgebra( L, H );
    SetRootSystem( L, R );
    pr := PositiveRoots( R );
    prv := PositiveRootVectors( R );
    nrv := NegativeRootVectors( R );
    cg := CanonicalGenerators( R );
    cbH := Basis( H, cg[3] );
    testIt := function (  )
          local C, basH, h, i, r, cf, cb;
          cb := ChevalleyBasis( R );
          C := LieCentraliser( L, H );
          if not IsAbelian( C ) or not Dimension( C ) = Dimension( H ) then
              Error( "not a CSA" );
          fi;
          if not cg[3] = cb[3]{[ 1 .. Length( cg[3] ) ]} then
              Error( "error 1" );
          fi;
          basH := Basis( H, cg[3] );
          for h in cg[3] do
              for i in [ 1 .. Length( pr ) ] do
                  r := pr[i] * Coefficients( basH, h );
                  cf 
                   := 
                    Coefficients( 
                       BasisNC( SubspaceNC( L, [ cb[1][i] ], "basis" ), 
                         [ cb[1][i] ] ), h * cb[1][i] )[1];
                  if not r = cf then
                      Error( "error cf" );
                  fi;
              od;
          od;
          return;
      end;
    decomposeCSA := function ( H, cbH )
          local theta, ev, esp, HK, HP, h, ad;
          theta := List( cbH, function ( x )
                  return Coefficients( cbH, - sigma( x ) );
              end );
          ev := [  ];
          ev[1] := List( NullspaceMat( theta - theta ^ 0 ), function ( x )
                  return x * cbH;
              end );
          ev[2] := List( NullspaceMat( theta + theta ^ 0 ), function ( x )
                  return x * cbH;
              end );
          if Length( ev[2] ) = 0 then
              Info( InfoCorelg, 3, "  now have CSA with compact dimension ", 
               Dimension( H ) );
              return rec(
                  basHK := cbH,
                  basHP := Basis( SubspaceNC( H, [  ], "basis" ) ) );
          elif Length( ev[1] ) = 0 then
              Info( InfoCorelg, 3, "  now have CSA with compact dimension 0" 
               );
              return rec(
                  basHP := cbH,
                  basHK := Basis( SubspaceNC( H, [  ], "basis" ) ) );
          fi;
          Info( InfoCorelg, 3, "  now have CSA with compact dimension ", 
           Length( ev[1] ) );
          return rec(
              basHK := ev[1],
              basHP := ev[2] );
      end;
    realRoot := function ( pr, basH, basHK )
          local alpha, real, h, cf;
          for alpha in pr do
              cf := List( basHK, function ( h )
                      return Coefficients( basH, h ) * alpha;
                  end );
              if ForAll( cf, function ( x )
                        return x = Zero( F );
                    end ) then
                  return alpha;
              fi;
          od;
          return false;
      end;
    decH := decomposeCSA( H, cbH );
    alpha := realRoot( pr, cbH, decH.basHK );
    while not alpha = false do
        i := Position( pr, alpha );
        x := prv[i];
        sx := sigma( x );
        if sx <> x then
            if sx <> - x then
                x := x + sx;
            else
                x := E( 4 ) * One( F ) * x;
            fi;
        fi;
        y := nrv[i];
        cf := Coefficients( Basis( SubspaceNC( L, [ x ], "basis" ), [ x ] ), 
             x * y * x )[1];
        y := 2 / cf * y;
        if not x * y * x = 2 * x or not x * y * y = -2 * y then
            Error( "not triple" );
        fi;
        K := List( NullspaceMat( TransposedMat( [ alpha ] ) ), function ( x )
                return One( F ) * x * cbH;
            end );
        rts := [  ];
        spc := [  ];
        rrv := Concatenation( prv, nrv );
        rrr := Concatenation( pr, - pr );
        for i in [ 1 .. Length( rrv ) ] do
            rt := List( K, function ( h )
                    return Coefficients( cbH, h ) * rrr[i];
                end );
            pos := Position( rts, rt );
            if pos = fail then
                Add( rts, rt );
                Add( spc, [ rrv[i] ] );
            else
                Add( spc[pos], rrv[i] );
            fi;
        od;
        pos := Position( rts, List( K, function ( h )
                  return 0;
              end ) );
        Append( spc[pos], Basis( H ) );
        H := SubspaceNC( L, Concatenation( K, [ x - y ] ), "basis" );
        R := corelg.specialrtsys( L, H, spc, K, x - y );
        pr := PositiveRoots( R );
        prv := PositiveRootVectors( R );
        nrv := NegativeRootVectors( R );
        cg := CanonicalGenerators( R );
        cbH := Basis( H, cg[3] );
        decH := decomposeCSA( H, cbH );
        alpha := realRoot( pr, cbH, decH.basHK );
    od;
    Info( InfoCorelg, 1, 
     "end MaximallyCompactCartanSubalgebra; found CSA with compact dim ", 
     Length( decH.basHK ) );
    SetcorelgCompactDimOfCSA( H, Length( decH.basHK ) );
        SetMaximallyCompactCartanSubalgebra( L, H );
    return H;
end;


diagmat:= function( F, H ) # F : field
                           # H torus, find a matrix C such that 
                           # C^-1*u*C is diagonal for all u in H


    local spaces, i, h, sp0, s, sp, A, fct, f, V, bas, C, one,
          fct0, num, fam, cf, b, c, r, n, U;

    one := One(F);
    n:= Length( H[1] );

    if ForAll( H, IsDiagonalMat ) then # nothing to do..
       return IdentityMat(n)*one;
    fi;

    V:= F^n;
    spaces:= [ ShallowCopy( BasisVectors( Basis(V) ) ) ];
    for i in [1..Length(H)] do
        h:= H[i];
        sp0:= [ ];
        for s in spaces do
            sp:= Basis( Subspace( V, s ), s );
            A:= List( s, u -> Coefficients( sp, h*u ) );
            fct:= Factors( MinimalPolynomial(F,A) );

            num  := IndeterminateNumberOfUnivariateLaurentPolynomial(fct[1]);
            fam  := FamilyObj( fct[1] );

            fct0:= [ ];
            for f in fct do
               if Degree(f) = 1 then
                  Add( fct0, f );
               elif Degree(f) = 2 then # we just take square roots...
                  cf := CoefficientsOfUnivariatePolynomial(f);
                  b  := cf[2]; 
                  c  := cf[1];
                  if F=SqrtField then
                     r  := (-b+Sqroot(b^2-4*c))/2;
                  else
                     r  := (-b+Sqrt(b^2-4*c))/2;
                  fi;
                  if not r in F then Print("diagmat: cannot do this over ",F,"\n"); fi;
                  Add( fct0, PolynomialByExtRep( fam, 
                                            [ [], -r, [num,1], one] ) );
                  if F=SqrtField then
                     r  := (-b-Sqroot(b^2-4*c))/2;
                  else
                     r  := (-b-Sqrt(b^2-4*c))/2;
                  fi;
                  if not r in F then Print("diagmat: cannot do this over ",F,"\n"); fi;
                  Add( fct0, PolynomialByExtRep( fam, 
                           [ [], -r, [num,1], one] ) );
               else
                  Error("not split!");
                  return fail;
               fi;
            od;

            for f in fct0 do
                U:= NullspaceMat( Value( f, A ) );
                Add( sp0, List( U, x -> LinearCombination( s, x ) ) );
            od;
        od;
        spaces:= sp0;
    od;
   
    bas:= Concatenation( spaces );
    C:= List( bas, x -> Coefficients( Basis(V), x ) );
    return TransposedMat(C); 

end;


toruspar_gen:= function( F, H ) # H list of mats giving the Lie algebra of torus
                             # output: parametrization of torus 

    local C, ad, add, diag, lat, n, A, d, r, S, P, Q, names, i, j, R, 
          indet, mats, m;

    n:= Length( H[1] );

    C:= diagmat(F,H);
    add:= List( H, u -> C^-1*u*C );

    diag:= List( add, a -> List( [1..n], i -> a[i][i] ) );
    lat:= NullspaceMat( TransposedMat(diag) );

    if lat = [ ] then lat:= [ List( [1..n], i -> Zero(F) ) ]; fi;

    # saturation...
    A:= lat;
    if F = SqrtField then
       A:= SqrtFieldMakeRational(A);
    fi;
    if A <> 0*A then
       d:= Lcm( List( Filtered( Flat(A), x -> not IsZero(x)), DenominatorRat));
       A:= d*A;
    fi;
    r:=SmithNormalFormIntegerMatTransforms(A);
    Q:= r.coltrans; P:= r.rowtrans;
    S:= List( r.normal, ShallowCopy );
    for i in [1..Length(S)] do 
        if S[i][i] > 0 then S[i][i]:= 1; fi;
    od;
    lat:= P^-1*S*Q^-1;

    lat:= NullspaceMat( TransposedMat(lat) );
    # another saturation...
    A:= lat;
    d:= Lcm( List( Filtered( Flat(A), x -> not IsZero(x) ), DenominatorRat ) );
    A:= d*A;
    r:=SmithNormalFormIntegerMatTransforms(A);
    Q:= r.coltrans; P:= r.rowtrans;
    S:= List( r.normal, ShallowCopy );
    for i in [1..Length(S)] do 
        if S[i][i] > 0 then S[i][i]:= 1; fi;
    od;
    lat:= P^-1*S*Q^-1;

    names:= [ ];
    for i in [1..Length(lat)] do
        Add( names, Concatenation( "a", String(i) ) );
    od;
    R:= PolynomialRing( F, names );
    
    indet:= IndeterminatesOfPolynomialRing( R );
    mats:=  [ ];
    for i in [1..Length(lat)] do
        m:= NullMat( n, n, R );
        for j in [1..n] do
            m[j][j]:= indet[i]^lat[i][j];
        od;
        #Add( mats, C*m*C^-1 );
        Add( mats, m );
    od;

    A:= TransposedMat(lat);
    Q:= HermiteNormalFormIntegerMatTransform( A ).rowtrans;

    return rec( mats:= mats, C:= C, indets:= indet, exps:=A, Q:= Q );
    

end;


comconj:= function( pol )

        # pol is a polynomial, we compute its complex conjugate

        local e, i;

        e:= ShallowCopy( ExtRepPolynomialRatFun(pol) );
        for i in [2,4..Length(e)] do
            e[i]:= ComplexConjugate( e[i] );
        od;
        return PolynomialByExtRep( FamilyObj(pol), e );
end;


valpars:= function( r, A )

        # here r is the output of torusparam_gen
        # A is a matrix in the torus (so the nonsplit torus)
        # with elements T(t1,...,tm)
        # we return a the vector [t1,..,tm] such that
        # T(t1,...,tm)=A

        local D, Q, t, m;

        D:= r.C^-1*A*r.C;
        if not IsDiagonalMat( D ) then return fail; fi;
        D:= List( [1..Length(D)], i -> D[i][i] );
        Q:= r.Q;
        t:= List( Q, s -> Product( List( [1..Length(s)], i -> D[i]^s[i] ) ) );
        m:= Length(r.exps[1]);
        if not ForAll([m+1..Length(t)], i -> IsOne(t[i]) ) then
           return fail; # the element does not lie in the torus...
        fi;
        return t{[1..m]};
end;


normalbasis:= function( tau )

        # tau is an involution of the lattice Z^m, we compute a
        # basis of the lattice with respect to which tau has diagonal form,
        # with 1, -1 and
        #
        #      0,1
        #      1,0
        #
        # on the diagonal...

        local m, ev, evals, i, pos, v1, d, k, Qi, B, taup, e, f, g, h, ff0,
              ff1, r;

        m:= Length(tau);
        if m=1 then
           if tau[1][1] = 1 then 
              return rec( e:= [ [1] ], f:= [], g:= [], h:= [] );
           else
              return rec( e:= [], f:= [ [1] ], g:= [], h:= [] );
           fi;
        fi;

        ev:= Eigenvectors( Rationals, TransposedMat(tau) );
        evals:= [ ];
        for i in [1..m] do
            if tau*ev[i] = ev[i] then
               Add( evals, 1 );
            elif tau*ev[i] = -ev[i] then
               Add( evals, -1 );
            else
               Print("something wrong here...\n");
            fi;
        od;

        pos:= Position( evals, 1 );

        if pos = fail then # only evals -1
           return rec( e:= [ ], f:= ev, g:= [ ], h:= [ ] );
        fi;

        v1:= ShallowCopy( ev[pos] );
        d:= Lcm( List( v1, x -> DenominatorRat( x ) ) ) ;
        v1:= v1*d;
        d:= Gcd(v1);
        v1:= v1/d;

        # now we find a basis of L = Z^m of the form v1, v2,...,vm.
        Qi:= ( SmithNormalFormIntegerMatTransforms( [ v1 ] ).coltrans )^-1;
        B:= Basis( Rationals^m, Qi );
        taup:= List( [2..m], k -> Coefficients(B,tau*Qi[k]){[2..m]} );
        taup:= TransposedMat(taup);

        r:= normalbasis( taup );
        e:= List( r.e, x -> x*Qi{[2..m]} );
        f:= List( r.f, x -> x*Qi{[2..m]} );
        g:= List( r.g, x -> x*Qi{[2..m]} );
        h:= List( r.h, x -> x*Qi{[2..m]} );

        for i in [1..Length(g)] do
            h[i]:= tau*g[i];
        od;

        B:= Basis( VectorSpace( Rationals, [v1] ), [v1] );
        ff0:= [ ]; # this collects the f with tau*f = -f
        ff1:= [ ]; # this collects the f with tau*f = -f + v1
        for i in [1..Length(f)] do
            k:= Coefficients( B, tau*f[i]+f[i] )[1];
            if IsEvenInt(k) then
               Add( ff0, f[i]-(k/2)*v1 );
            else
               Add( ff1, f[i]-((k-1)/2)*v1 );
            fi;
        od;

        for i in [2..Length(ff1)] do
            Add( ff0, ff1[i]-ff1[1] );
        od;

        if Length(ff1) > 0 then
           Add( g, ff1[1] );
           Add( h, tau*ff1[1] );
        else
           Add( e, v1 );
        fi;
        return rec( e:= e, f:= ff0, g:= g, h:= h ); 
end;

isomTS:= function( tau )

        local r, A, B, m, TtoS, StoT; 

        r:= normalbasis( tau );
        A:= Concatenation( r.f, r.e, r.g, r.h );
        A:= TransposedMat(A);
        B:= A^-1;

        m:= Length(tau);

        TtoS:= function( t )

              return List( [1..m], j -> Product( List([1..m], k ->
                        t[k]^B[j][k] ) ) );
        end;

        StoT:= function(s)
              return List( [1..m], j -> Product( List([1..m], k ->
                        s[k]^A[j][k] ) ) );
        end;        

        return rec( flen:= Length(r.f), elen:= Length(r.e),
            glen:= Length(r.g), hlen:= Length(r.h),             
            TtoS:= TtoS, StoT:= StoT );
end;


H1connected:= function( arg )

        # arg can consist of mats or of mats and QQ
        # here mats span a reductive matrix Lie algebra, we compute the
        # H1 of the corresponding complex group, where the conjugation is
        # given by g ---> QQ*\bar{g}*QQ^-1...

        local K, D, cd, t0, H, t, r, Q, A, U, diag, v, m, inds, tau, i, k, a,
              R, simref, cg, cl, ex, ey, wds, elms, w, H1H, reps, j, M, is,
              perms, g, list, u, orbs, pos, tup, tail, s, mats, QQ, SS, ggg,
              hhh, subg, PG, W0, Pr, Ci;

        mats:= arg[1];
        if Length(arg) = 2 then
           QQ:= arg[2];
        else
           QQ:= mats[1]^0;
        fi;

        K:= LieAlgebra( SqrtField, mats );
        D:= LieDerivedSubalgebra( K );
        SetRealStructure( D,
         function( v ) return QQ*LieObject(ComplexConjugate(v))*QQ^-1; end );
        if Dimension(D) > 0 then
           # the next statement is in itself useless as H will be computed anew,
           # but it is there not to run into trouble in CartanDecomposition,
           # where the max cpt CSA from CoReLG is called, which requires a
           # basis of a CSA to be invariant under sigma.
           H:= maxcptCSA( D );
           cd:= CartanDecomposition( D );
           t0:= CartanSubalgebra( cd.K );
           H:= LieCentralizer( K, t0 );
        else
           H:= K;
        fi;
        t:= Intersection( H, D );

        # we find the conjugation on the torus corr to H...
        r:= toruspar_gen( SqrtField, List( Basis(H), x -> x![1] ) );
        Q:= r.Q;
        A:= r.C^-1*QQ*ComplexConjugate( r.C );
        U:= A*Product( r.mats )*A^-1;
        diag:= List( [1..Length(U)], i -> U[i][i] );
        v:= List( Q, s -> Product( List( [1..Length(s)], i -> diag[i]^s[i])));

        m:= Dimension(H);
        inds:= List( r.indets, x -> ExtRepPolynomialRatFun(x)[1][1] );
        tau:= NullMat(m,m);
        for i in [1..m] do
            a:= ExtRepNumeratorRatFun( v[i] );
            for k in [1,3..Length(a[1])-1] do
                pos:= Position( inds, a[1][k] );
                tau[i][pos]:= tau[i][pos]+a[1][k+1];
            od;
            a:= ExtRepDenominatorRatFun( v[i] );
            for k in [1,3..Length(a[1])-1] do
                pos:= Position( inds, a[1][k] );
                tau[i][pos]:= tau[i][pos]-a[1][k+1];
            od;
        od;


        # now compute the H1 of the H
        is:= isomTS( tau );

        # compute the subtori on the "S"-side...
        SS:= [ ];
        for i in [1..Length(r.indets)] do
            v:= List( r.indets, x -> 1 );
            v[i]:= r.indets[i];
            v:= is.StoT( v );
            M:= r.C*Product( r.mats )*r.C^-1;
            for i in [1..Length(M)] do
                for j in [1..Length(M)] do
                    M[i][j]:= Value( M[i][j], r.indets, v );
                od;
            od;
            Add( SS, M );
        od;

        
        tup:= -Tuples( [-1,1], is.flen );
        tail:= List( [is.flen+1..Length(tau)], x -> 1 );
        H1H:= List( tup, x -> is.StoT(Concatenation( x, tail )) );

        #reps:= [ ];
        #for k in [1..Length(H1H)] do
        #    M:= r.C*Product( r.mats )*r.C^-1;
        #    for i in [1..Length(M)] do
        #        for j in [1..Length(M)] do
        #            M[i][j]:= Value( M[i][j], r.indets, H1H[k] );
        #        od;
        #    od;
        #    Add( reps, M );
        #od;

        reps:= [ ];
        Pr:= Product( r.mats );
        Ci:= r.C^-1;
        for k in [1..Length(H1H)] do
            M:= List( Pr, ShallowCopy );
            for i in [1..Length(M)] do
                for j in [1..Length(M)] do
                    M[i][j]:= Value( M[i][j], r.indets, H1H[k] );
                od;
            od;
            Add( reps, r.C*M*Ci );
        od;


        if Dimension(D) = 0 then
           return rec( cocyc:= reps, torusdata:=r, is:= is,
                       H1S:= List( H1H, x -> is.TtoS(x) ), QQ:= QQ,
                       SS:=SS);
        fi;


        # Now we compute the root system wrt t, its Weyl group and the
        # subgroup of elements stabilizing t0

        R:= RootsystemOfCartanSubalgebra( D, t );
        W0:= RealWeylGroup( D, t );

        # we compute elements in the group corr to our Lie algebra
        # inducing the simple reflections

        simref:= [ ];
        cg:= CanonicalGenerators( R );
        cg:= [ List(cg[1],x -> x![1]), List( cg[2], x -> x![1] ) ];
        for i in [1..Length(cg[1])] do
            cl:= 1;
            while not IsZero( cg[1][i]^cl ) do cl:= cl+1; od;
            ex:= Sum([0..cl-1], j -> cg[1][i]^j/Factorial(j) );
            cl:= 1;
            while not IsZero( cg[2][i]^cl ) do cl:= cl+1; od;
            ey:= Sum([0..cl-1], j -> (-cg[2][i])^j/Factorial(j) );
            Add( simref, ex*ey*ex );
        od;


if printstuff then 
   Print("H1connected: start computing subgroup of Weyl group...\n");
fi;
        ggg:= RightTransversal(WeylGroupAsPermGroup(R),W0);
        hhh:= [];
        subg:= W0;
        wds:= List( ggg, u -> PermAsWeylWord( R, u ) );
        elms:= List( GeneratorsOfGroup(W0), g ->
                    Product( simref{PermAsWeylWord(R,g)} ) );
        for i in [1..Length(wds)] do
            w:= wds[i];
            a:= Product( simref{w} );
            if ForAll( Basis(t0), x -> a*x*a^-1 in t0 ) then
               Add( elms, a );
               #Add( hhh, ggg[i] );
               #subg:= Group( hhh );
            fi;
        od;

if printstuff then 
   Print("H1connected: done subgroup...\n");
fi;
        perms:= [ ];
        for g in elms do
            list:= [ ];
            for i in [1..Length(reps)] do
                U:= g^-1*reps[i]*(QQ*ComplexConjugate(g)*QQ^-1);
                u:= valpars( r, U );
                s:= is.TtoS( u );
                for i in [1..is.flen] do
                    if s[i] > 0 then s[i]:=1; fi;
                    if s[i] < 0 then s[i]:= -1; fi;
                od;
                for i in [is.flen+1..Length(s)] do
                    s[i]:=1;
                od;
                Add( list, Position( H1H, is.StoT(s) ) );
            od;
            Add( perms, PermList( list ) );
        od;

        PG:= Group( perms );
        orbs:= Orbits( PG, [1..Length(H1H)], OnPoints );

        return rec( cocyc:= List( orbs, x -> reps[x[1]] ),
            L:= K, H:= H, is:= is, SS:= SS, torusdata:= r, W0:= elms,
            perms:= PG, H1S:= List( H1H, x -> is.TtoS(x) ),
            orbreps:= List( orbs, x -> x[1] ), QQ:=QQ );

end;

FirstGaloisCohomologyconnected:= function( arg )

        # this is the same as the previous function, except that we do it
        # over CF(4), if possible. If not we revert to the previous function.

        # arg can consist of mats or of mats and QQ
        # here mats span a reductive matrix Lie algebra, we compute the
        # H1 of the corresponding complex group, where the conjugation is
        # given by g ---> QQ*\bar{g}*QQ^-1...

        local K, D, cd, t0, H, t, r, Q, A, U, diag, v, m, inds, tau, i, k, a,
              R, simref, cg, cl, ex, ey, wds, elms, w, H1H, reps, j, M, is,
              perms, g, list, u, orbs, pos, tup, tail, s, mats, QQ, SS, ggg,
              hhh, subg, PG, W0, F, Pr, Ci, t1, t2;

        mats:= arg[1];
        if Length(arg) = 2 then
           QQ:= arg[2];
        else
           QQ:= mats[1]^0;
        fi;

        if DefaultFieldOfMatrix( mats[1] ) = SqrtField then
           r:= H1connected(arg);
           return r.cocyc;
        fi;

        elms:= Flat(mats); Add( elms, E(4) );
        F:= DefaultFieldOfMatrix( [elms] );
        
t1:= Runtime();
        K:= LieAlgebra( F, mats );
        D:= LieDerivedSubalgebra( K );
        SetRealStructure( D,
         function( v ) return QQ*LieObject(ComplexConjugate(v))*QQ^-1; end );
        if Dimension(D) > 0 then
           # the next statement is in itself useless as H will be computed anew,
           # but it is there not to run into trouble in CartanDecomposition,
           # where the max cpt CSA from CoReLG is called, which requires a
           # basis of a CSA to be invariant under sigma.
           H:= maxcptCSA( D );
           cd:= CartanDecomposition( D );
           t0:= CartanSubalgebra( cd.K );
           H:= LieCentralizer( K, t0 );
        else
           H:= K;
        fi;
        t:= Intersection( H, D );
Print("done initialization....\n");
Print( Runtime()-t1,"\n");
t1:= Runtime();
        # we find the conjugation on the torus corr to H...
        r:= toruspar_gen( F, List( Basis(H), x -> x![1] ) );
        Pr:= Product( r.mats );
        Ci:= r.C^-1;
        Q:= r.Q;
        A:= Ci*QQ*ComplexConjugate( r.C );
        U:= A*Pr*A^-1;
        diag:= List( [1..Length(U)], i -> U[i][i] );
        v:= List( Q, s -> Product( List( [1..Length(s)], i -> diag[i]^s[i])));

        m:= Dimension(H);
        inds:= List( r.indets, x -> ExtRepPolynomialRatFun(x)[1][1] );
        tau:= NullMat(m,m);
        for i in [1..m] do
            a:= ExtRepNumeratorRatFun( v[i] );
            for k in [1,3..Length(a[1])-1] do
                pos:= Position( inds, a[1][k] );
                tau[i][pos]:= tau[i][pos]+a[1][k+1];
            od;
            a:= ExtRepDenominatorRatFun( v[i] );
            for k in [1,3..Length(a[1])-1] do
                pos:= Position( inds, a[1][k] );
                tau[i][pos]:= tau[i][pos]-a[1][k+1];
            od;
        od;
Print("tau....  \n",tau,"\n\n");
Print( Runtime()-t1,"\n");
t1:= Runtime();

        # now compute the H1 of the H
        is:= isomTS( tau );

        # compute the subtori on the "S"-side...
        SS:= [ ];
        for i in [1..Length(r.indets)] do
            v:= List( r.indets, x -> 1 );
            v[i]:= r.indets[i];
            v:= is.StoT( v );
            #M:= r.C*Pr*r.C^-1;
            M:= List( Pr, ShallowCopy );
            for i in [1..Length(M)] do
                for j in [1..Length(M)] do
                    M[i][j]:= Value( M[i][j], r.indets, v );
                od;
            od;
            Add( SS, r.C*M*Ci );
        od;

Print("computed S...\n");
Print( Runtime()-t1,"\n");
t1:= Runtime();

        tup:= -Tuples( [-1,1], is.flen );
        tail:= List( [is.flen+1..Length(tau)], x -> 1 );
        H1H:= List( tup, x -> is.StoT(Concatenation( x, tail )) );

Print("computed H1H.... ",Length(H1H),"\n");
Print( Runtime()-t1,"\n");
t1:= Runtime();

        reps:= [ ];
        for k in [1..Length(H1H)] do
            #M:= r.C*Pr*Ci;
            M:= List( Pr, ShallowCopy );
            for i in [1..Length(M)] do
                for j in [1..Length(M)] do
                    M[i][j]:= Value( M[i][j], r.indets, H1H[k] );
                od;
            od;
            Add( reps, r.C*M*Ci );
        od;
Print("H1T computed....\n");
Print( Runtime()-t1,"\n");
t1:= Runtime();
        if Dimension(D) = 0 then
           return reps;
        fi;


        # Now we compute the root system wrt t, its Weyl group and the
        # subgroup of elements stabilizing t0

        R:= RootsystemOfCartanSubalgebra( D, t );
        W0:= RealWeylGroup( D, t );

        # we compute elements in the group corr to our Lie algebra
        # inducing the simple reflections

Print("start computing Weyl orbits...\n");
Print( Runtime()-t1,"\n");
t1:= Runtime();
        simref:= [ ];
        cg:= CanonicalGenerators( R );
        cg:= [ List(cg[1],x -> x![1]), List( cg[2], x -> x![1] ) ];
        for i in [1..Length(cg[1])] do
            cl:= 1;
            while not IsZero( cg[1][i]^cl ) do cl:= cl+1; od;
            ex:= Sum([0..cl-1], j -> cg[1][i]^j/Factorial(j) );
            cl:= 1;
            while not IsZero( cg[2][i]^cl ) do cl:= cl+1; od;
            ey:= Sum([0..cl-1], j -> (-cg[2][i])^j/Factorial(j) );
            Add( simref, ex*ey*ex );
        od;


if printstuff then 
   Print("H1connected: start computing subgroup of Weyl group...\n");
fi;
Print( Runtime()-t1,"\n");
t1:= Runtime();
        ggg:= RightTransversal(WeylGroupAsPermGroup(R),W0);
        hhh:= ShallowCopy( GeneratorsOfGroup(W0) );
        subg:= W0;
        wds:= List( ggg, u -> PermAsWeylWord( R, u ) );
        elms:= List( GeneratorsOfGroup(W0), g ->
                    Product( simref{PermAsWeylWord(R,g)} ) );
        for i in [1..Length(wds)] do
            if not ggg[i] in subg then
               w:= wds[i];
               a:= Product( simref{w} );
               if ForAll( Basis(t0), x -> a*x*a^-1 in t0 ) then
                  Add( elms, a );
                  Add( hhh, ggg[i] );
                  subg:= Group( hhh );
               fi;
            fi;
       od;

if printstuff then 
   Print("H1connected: done subgroup...\n");
#Print("Order of W0.... ",Size(W0),"\n");
fi;
Print("len elms... ",Length(elms)," len reps... ",Length(reps),"\n");
Print( Runtime()-t1,"\n");
t1:= Runtime();
        perms:= [ ];
        for g in elms do
            list:= [ ];
            for i in [1..Length(reps)] do
#Print(i,"  ",valpars(r,reps[i]),"\n");
                U:= g^-1*reps[i]*(QQ*ComplexConjugate(g)*QQ^-1);
#Print(r.C^-1*U*r.C,"\n\n");
                u:= valpars( r, U );
#Print(u,"\n");
                s:= is.TtoS( u );
                for i in [1..is.flen] do
                    if s[i] > 0 then s[i]:=1; fi;
                    if s[i] < 0 then s[i]:= -1; fi;
                od;
                for i in [is.flen+1..Length(s)] do
                    s[i]:=1;
                od;
#Print(is.StoT(s),"\n");
                Add( list, Position( H1H, is.StoT(s) ) );
            od;
            Add( perms, PermList( list ) );
        od;

#Print(perms,"\n",H1H,"\n");
Print("perms computed....\n");
Print( Runtime()-t1,"\n");
t1:= Runtime();

        PG:= Group( perms );
        orbs:= Orbits( PG, [1..Length(H1H)], OnPoints );

        return List( orbs, x -> reps[x[1]] );

end;


H1fast:= function( L, V )

        # here L is a semisimple Lie algebra over a field of characteristic
        # 0, such as Rationals or CF(4), not SqrtField.
        # V is an L-module, we compute the first Galois cohomology of the
        # corresponding group in GL(V). 

        # this is the same as the previous function, except that we do it
        # over CF(4), if possible. If not we revert to the previous function.

        # arg can consist of mats or of mats and QQ
        # here mats span a reductive matrix Lie algebra, we compute the
        # H1 of the corresponding complex group, where the conjugation is
        # given by g ---> QQ*\bar{g}*QQ^-1...

        local K, D, cd, t0, H, t, r, Q, A, U, diag, v, m, inds, tau, i, k, a,
              R, simref, cg, cl, ex, ey, wds, elms, w, H1H, reps, j, M, is,
              perms, g, list, u, orbs, pos, tup, tail, s, mats, QQ, SS, ggg,
              hhh, subg, PG, W0, F, Pr, Ci, t1, t2;

        F:= CF(4);
        QQ:= IdentityMat( Dimension(V) );
        H:= MaximallyCompactCartanSubalgebra(L);
        # we find the conjugation on the torus corr to H...
        r:= toruspar_gen( F, List( Basis(H), x -> MatrixOfAction(Basis(V),x)) );
        Pr:= Product( r.mats );
        Ci:= r.C^-1;
        Q:= r.Q;
        A:= Ci*QQ*ComplexConjugate( r.C );
        U:= A*Pr*A^-1;
        diag:= List( [1..Length(U)], i -> U[i][i] );
        v:= List( Q, s -> Product( List( [1..Length(s)], i -> diag[i]^s[i])));

        m:= Dimension(H);
        inds:= List( r.indets, x -> ExtRepPolynomialRatFun(x)[1][1] );
        tau:= NullMat(m,m);
        for i in [1..m] do
            a:= ExtRepNumeratorRatFun( v[i] );
            for k in [1,3..Length(a[1])-1] do
                pos:= Position( inds, a[1][k] );
                tau[i][pos]:= tau[i][pos]+a[1][k+1];
            od;
            a:= ExtRepDenominatorRatFun( v[i] );
            for k in [1,3..Length(a[1])-1] do
                pos:= Position( inds, a[1][k] );
                tau[i][pos]:= tau[i][pos]-a[1][k+1];
            od;
        od;

        # now compute the H1 of the H
        is:= isomTS( tau );

        # compute the subtori on the "S"-side...
        SS:= [ ];
        for i in [1..Length(r.indets)] do
            v:= List( r.indets, x -> 1 );
            v[i]:= r.indets[i];
            v:= is.StoT( v );
            #M:= r.C*Pr*r.C^-1;
            M:= List( Pr, ShallowCopy );
            for i in [1..Length(M)] do
                for j in [1..Length(M)] do
                    M[i][j]:= Value( M[i][j], r.indets, v );
                od;
            od;
            Add( SS, r.C*M*Ci );
        od;

        tup:= -Tuples( [-1,1], is.flen );
        tail:= List( [is.flen+1..Length(tau)], x -> 1 );
        H1H:= List( tup, x -> is.StoT(Concatenation( x, tail )) );

        reps:= [ ];
        for k in [1..Length(H1H)] do
            #M:= r.C*Pr*Ci;
            M:= List( Pr, ShallowCopy );
            for i in [1..Length(M)] do
                for j in [1..Length(M)] do
                    M[i][j]:= Value( M[i][j], r.indets, H1H[k] );
                od;
            od;
            Add( reps, r.C*M*Ci );
        od;

        # Now we compute the root system wrt t, its Weyl group and the
        # subgroup of elements stabilizing t0

        R:= RootsystemOfCartanSubalgebra( L, H );
        W0:= RealWeylGroup( L, H );

        # we compute elements in the group corr to our Lie algebra
        # inducing the simple reflections

        simref:= [ ];
        cg:= CanonicalGenerators( R );
        cg:= [ List( cg[1], x -> MatrixOfAction(Basis(V),x)),
               List( cg[2], x -> MatrixOfAction(Basis(V),x)) ];
        for i in [1..Length(cg[1])] do
            cl:= 1;
            while not IsZero( cg[1][i]^cl ) do cl:= cl+1; od;
            ex:= Sum([0..cl-1], j -> cg[1][i]^j/Factorial(j) );
            cl:= 1;
            while not IsZero( cg[2][i]^cl ) do cl:= cl+1; od;
            ey:= Sum([0..cl-1], j -> (-cg[2][i])^j/Factorial(j) );
            Add( simref, ex*ey*ex );
        od;

        ggg:= RightTransversal(WeylGroupAsPermGroup(R),W0);
        hhh:= ShallowCopy( GeneratorsOfGroup(W0) );
        subg:= W0;
        wds:= List( ggg, u -> PermAsWeylWord( R, u ) );
        elms:= List( GeneratorsOfGroup(W0), g ->
                    Product( simref{PermAsWeylWord(R,g)} ) );
        t0:= List( Basis( Intersection( H, CartanDecomposition(L).K ) ),
                   x-> MatrixOfAction( Basis(V), x ) );
        t0:= VectorSpace( LeftActingDomain(L), t0 );
        for i in [1..Length(wds)] do
            if not ggg[i] in subg then
               w:= wds[i];
               a:= Product( simref{w} );
               if ForAll( Basis(t0), x -> a*x*a^-1 in t0 ) then
                  Add( elms, a );
                  Add( hhh, ggg[i] );
                  subg:= Group( hhh );
               fi;
            fi;
       od;

        perms:= [ ];
        for g in elms do
            list:= [ ];
            for i in [1..Length(reps)] do
                U:= g^-1*reps[i]*(QQ*ComplexConjugate(g)*QQ^-1);
                u:= valpars( r, U );
                s:= is.TtoS( u );
                for i in [1..is.flen] do
                    if s[i] > 0 then s[i]:=1; fi;
                    if s[i] < 0 then s[i]:= -1; fi;
                od;
                for i in [is.flen+1..Length(s)] do
                    s[i]:=1;
                od;
                Add( list, Position( H1H, is.StoT(s) ) );
            od;
            Add( perms, PermList( list ) );
        od;

        PG:= Group( perms );
        orbs:= Orbits( PG, [1..Length(H1H)], OnPoints );

        return List( orbs, x -> reps[x[1]] );

end;



value:= function( pol, vars, vals )


        if IsCyc(pol) then
           return pol*One(SqrtField);
        elif IsSqrtFieldElement(pol) then
           return pol;
        else
           return Value( pol, vars, vals )*One(SqrtField);
        fi;
end;


normalize_toruscoc:= function( s, c )

        # s is the output of H1connected, c is a cocycle in the torus
        # we find a g in G such that g^-1*c*\bar{g} is one of the cocycles
        # in s.cocyc.

        local t, ts, i, k, l, pos, g1, c1, gg, u, v, g, ub, a, b, onepol,
              j, pol, epol;

        onepol:= s.torusdata.indets[1]^0;

        t:= valpars( s.torusdata, c );

        ts:= s.is.TtoS( t );

        for i in [1..s.is.flen] do
            if ts[i] > 0 then ts[i]:=1; fi;
            if ts[i] < 0 then ts[i]:= -1; fi;
        od;
        for i in [s.is.flen+1..Length(ts)] do
            ts[i]:=1;
        od;

        i:= Position( s.H1S, ts );
        if IsBound( s.orbreps ) and not i in s.orbreps then 
           gg:= List( s.orbreps, x -> RepresentativeAction( s.perms, i, x ) );
           pos:= PositionProperty( gg, x -> x <> fail );
           pol:= Factorization( s.perms, gg[pos] );
           epol:= ExtRepOfObj(pol);
           g1:= s.W0[1]^0;
           for j in [1,3..Length(epol)-1] do
               g1:= g1*s.W0[epol[j]]^epol[j+1];
           od;
           c1:= g1^-1*c*(s.QQ*ComplexConjugate(g1)*s.QQ^-1);
        else
           g1:= s.QQ^0;
           c1:= c;
        fi;

        t:= valpars( s.torusdata, c1 );
        ts:= s.is.TtoS( t );

        gg:= [ ];
        for i in [1..s.is.flen] do
            u:= ts[i];
            if ispositive(u) = true then
               v:= Sqroot(u);
            elif ispositive(u) = false then
               v:= Sqroot(-u);
            else
               Print("ERROR!! Cannot determine whether ",u," is pos or neg.\n");
            fi;
            g:= List( s.SS[i], ShallowCopy );
            for k in [1..Length(g)] do
                for l in [1..Length(g)] do
                    g[k][l]:= value( g[k][l], [ s.torusdata.indets[i] ], [v] );
                od;
            od;
            Add( gg, g );
        od;

        for i in [s.is.flen+1..s.is.flen+s.is.elen] do
            u:= ts[i];
            if IsOne(u) then
               v:= u;
            else
               ub:= ComplexConjugate(u);
               a:= (u+ub)/2;
               b:= (u-ub)/(2*E(4));
               v:= b+(-a+1)*E(4);
            fi;
            g:= List( s.SS[i], ShallowCopy );
            for k in [1..Length(g)] do
                for l in [1..Length(g)] do
                    g[k][l]:= value( g[k][l], [ s.torusdata.indets[i] ], [v] );
                od;
            od;
            Add( gg, g );
        od;

        for i in [s.is.flen+s.is.elen+1..s.is.flen+s.is.elen+s.is.glen] do
            v:= ts[i];
            g:= List( s.SS[i], ShallowCopy );
            for k in [1..Length(g)] do
                for l in [1..Length(g)] do
                    g[k][l]:= value( g[k][l], [ s.torusdata.indets[i] ], [v] );
                od;
            od;
            Add( gg, g );
        od;

        return g1*Product(gg);
end;


H1finite:= function( tab, gam )


        # tab is an nxn matrix with in each row and each column the elements
        # 1,...,n. gam is a function taking as input i in {1..n} and giving the
        # complex conjugate of the i-th element


        local inv, prod, n, coc, classes, dones, sets, pos, class, one;

        n:= Length( tab );
        one:= Position( tab, [1..n] );

        inv:= function( i )
               return Position( tab[i], one );
        end;

        prod:= function( i, j )
              return tab[i][j];
        end;



        coc:= Filtered( [1..n], x -> prod( x, gam(x) ) = one );
        classes:= [ ];
        dones:= [ ];
        sets:= [ ];
        while Length( dones ) < Length( coc ) do
            pos:= PositionProperty( coc, x -> not x in dones );
            class:= Set( List( [1..n], b ->
                      prod( prod( b, coc[pos] ), gam(inv(b)) ) ) );

            if one in class then
               Add( classes, one );
            else
               # see if there is a real element...
               pos:= PositionProperty( class, x -> x = gam(x) );
               if pos <> fail then
                  Add( classes, class[pos] );
               else
                  Add( classes, class[1] );
               fi;
            fi;
            Append( dones, Set(class) );
        od;

        return classes;

end;


maxcpttorus:= function( L, H0 )

    # here we find a maximally compact CSA of L that contains
    # the compact part of a given CSA H0    
        
    local F, sigma, R, pr, cb, cg, cbH, testIt, decomposeCSA, realRoot, 
    decH, alpha, i, x, y, j, K, prv, nrv, ev, rts, rt, rrr, rrv, spc, pos, 
    cf, sx, H;

    F := LeftActingDomain( L );
    sigma := RealStructure( L );

    H:= H0;

    R := RootsystemOfCartanSubalgebra( L, H );
    pr := PositiveRoots( R );
    prv := PositiveRootVectors( R );
    nrv := NegativeRootVectors( R );
    cg := CanonicalGenerators( R );
    cbH := Basis( H, cg[3] );
    testIt := function (  )
          local C, basH, h, i, r, cf, cb;
          cb := ChevalleyBasis( R );
          C := LieCentraliser( L, H );
          if not IsAbelian( C ) or not Dimension( C ) = Dimension( H ) then
              Error( "not a CSA" );
          fi;
          if not cg[3] = cb[3]{[ 1 .. Length( cg[3] ) ]} then
              Error( "error 1" );
          fi;
          basH := Basis( H, cg[3] );
          for h in cg[3] do
              for i in [ 1 .. Length( pr ) ] do
                  r := pr[i] * Coefficients( basH, h );
                  cf 
                   := 
                    Coefficients( 
                       BasisNC( SubspaceNC( L, [ cb[1][i] ], "basis" ), 
                         [ cb[1][i] ] ), h * cb[1][i] )[1];
                  if not r = cf then
                      Error( "error cf" );
                  fi;
              od;
          od;
          Print( "test ok\n" );
          return;
      end;
    decomposeCSA := function ( H, cbH )
          local theta, ev, esp, HK, HP, h, ad;
          theta := List( cbH, function ( x )
                  return Coefficients( cbH, - sigma( x ) );
              end );
          ev := [  ];
          ev[1] := List( NullspaceMat( theta - theta ^ 0 ), function ( x )
                  return x * cbH;
              end );
          ev[2] := List( NullspaceMat( theta + theta ^ 0 ), function ( x )
                  return x * cbH;
              end );
          if Length( ev[2] ) = 0 then
              Info( InfoCorelg, 3, "  now have CSA with compact dimension ", 
               Dimension( H ) );
              return rec(
                  basHK := cbH,
                  basHP := Basis( SubspaceNC( H, [  ], "basis" ) ) );
          elif Length( ev[1] ) = 0 then
              Info( InfoCorelg, 3, "  now have CSA with compact dimension 0" 
               );
              return rec(
                  basHP := cbH,
                  basHK := Basis( SubspaceNC( H, [  ], "basis" ) ) );
          fi;
          Info( InfoCorelg, 3, "  now have CSA with compact dimension ", 
           Length( ev[1] ) );
          return rec(
              basHK := ev[1],
              basHP := ev[2] );
      end;
    realRoot := function ( pr, basH, basHK )
          local alpha, real, h, cf;
          for alpha in pr do
              cf := List( basHK, function ( h )
                      return Coefficients( basH, h ) * alpha;
                  end );
              if ForAll( cf, function ( x )
                        return x = Zero( F );
                    end ) then
                  return alpha;
              fi;
          od;
          return false;
      end;
    decH := decomposeCSA( H, cbH );
    alpha := realRoot( pr, cbH, decH.basHK );
    while not alpha = false do
        i := Position( pr, alpha );
        x := prv[i];
        sx := sigma( x );
        if sx <> x then
            if sx <> - x then
                x := x + sx;
            else
                x := E( 4 ) * One( F ) * x;
            fi;
        fi;
        y := nrv[i];
        cf := Coefficients( Basis( SubspaceNC( L, [ x ], "basis" ), [ x ] ), 
             x * y * x )[1];
        y := 2 / cf * y;
        if not x * y * x = 2 * x or not x * y * y = -2 * y then
            Error( "not triple" );
        fi;
        K := List( NullspaceMat( TransposedMat( [ alpha ] ) ), function ( x )
                return One( F ) * x * cbH;
            end );
        rts := [  ];
        spc := [  ];
        rrv := Concatenation( prv, nrv );
        rrr := Concatenation( pr, - pr );
        for i in [ 1 .. Length( rrv ) ] do
            rt := List( K, function ( h )
                    return Coefficients( cbH, h ) * rrr[i];
                end );
            pos := Position( rts, rt );
            if pos = fail then
                Add( rts, rt );
                Add( spc, [ rrv[i] ] );
            else
                Add( spc[pos], rrv[i] );
            fi;
        od;
        pos := Position( rts, List( K, function ( h )
                  return 0;
              end ) );
        Append( spc[pos], Basis( H ) );
        H := SubspaceNC( L, Concatenation( K, [ x - y ] ), "basis" );
        R := corelg.specialrtsys( L, H, spc, K, x - y );
        pr := PositiveRoots( R );
        prv := PositiveRootVectors( R );
        nrv := NegativeRootVectors( R );
        cg := CanonicalGenerators( R );
        cbH := Basis( H, cg[3] );
        decH := decomposeCSA( H, cbH );
        alpha := realRoot( pr, cbH, decH.basHK );
    od;

    SetcorelgCompactDimOfCSA( H, Length( decH.basHK ) );
    if not HasCartanSubalgebra( L ) then
        SetCartanSubalgebra( L, H );
    fi;
    return H;
end;


autom:= function( L, H1 )

        # L a semisimple real Lie algebra, H1 a max cpt CSA, we
        # construct an automorphism of L such that the standard
        # max cpt CSA of L is mapped to H1

        # we do it for every simple piece separately, as otherwise it can
        # happen that the resulting automorphism permutes isomorphic simple
        # pieces...

        local cfs1, T, F, L1, f, vv, sig, cfs2, sig0;

        cfs1:= List( Basis(H1), x -> Coefficients( Basis(L), x ) );
        T:= StructureConstantsTable( Basis(L) );
        F:= LeftActingDomain(L);
        L1:= LieAlgebraByStructureConstants( F, T );
        SetMaximallyCompactCartanSubalgebra( L1, Subalgebra( L1,
                   List( cfs1, x -> x*Basis(L1) ) ) );

        sig:= RealStructure( L );
        cfs2:= List( Basis(L), x -> Coefficients( Basis(L), sig(x) ) );
        sig0:= function( v )
         return (List( Coefficients( Basis(L1), v ), ComplexConjugate )*cfs2)*
                  Basis(L1);
        end;
        SetRealStructure( L1, sig0 );

        f:= IsomorphismOfRealSemisimpleLieAlgebras( L, L1 );

        vv:= List( Basis(L), x -> Coefficients(Basis(L1),Image(f,x))*Basis(L) );
        f:= AlgebraHomomorphismByImagesNC( L, L, Basis(L), vv );

        return f;
end;



# Here we have a function for deciding whether an automorphism of a
# semisimple Lie algebra is inner. We use the algorithm of Cohen, Murray,
# Taylor, for the specific case of the adjoint rep.

weylwordNF:= function( R, path )     

        local   w,  rho,  wt;     

# the WeylWord in lex shortest form...

       w:= WeylWord( path );
       rho:= List( [1..Length( CartanMatrix(R))], x -> 1 );
       wt:= ApplyWeylElement( WeylGroup(R), rho, w );
       return ConjugateDominantWeightWithWord( WeylGroup(R), wt )[2];

end;

mons0:= function( R, hw )

        # R a root system, hw a highest weight
        # we get monomials in the f-s that applied to the highest weight
        # vector give a basis of the weight spaces 

        local cg, strings, paths, p, p1, st, word, j, k;

        cg:= CrystalGraph( R, hw );
        paths:=  cg.points; #Filtered( cg.points, p -> IsZero( EndWeight(p) ) );
        
        # For every path we compute its adapted string.
        strings:= [ ];
        for p in paths do
            p1:= p;
            st:= [];
            word:= weylwordNF( R, p1 );
            while word <> [] do
                j:= 0;
                k:= word[1];
                while  WeylWord( p1 ) <> [] and word[1] = k do
                     p1:= Ealpha( p1, k );
                     word:= weylwordNF( R, p1 );     
                     j:= j+1;
                od;
                if j > 0 then
                    Add( st, k );
                    Add( st, j );
                fi;
            od;
            Add( strings, st );        
        od;

        return strings;

end;

conformmat:= function( L, R, ch, BL )

        # L: Lie alg,
        # R: its root system,
        # ch: corr Chevalley basis
        # BL: a basis of L...

        # the matrix of the contravariant form of the adjoint rep of L
        # restricted to the CSA.

        local N, hw, st, v0, f, e, bash, s, u, n, j, k, mat, sp, i, l, A, cf;

        N:= Length( PositiveRoots(R) );
        hw:= PositiveRootsAsWeights( R )[N];
        st:= mons0( R, hw );

        v0:= ch[1][N];
        f:= CanonicalGenerators(R)[2];
        e:= CanonicalGenerators(R)[1];

        bash:= [ ];
        for s in st do

            u:= v0;
            n:= Length(s);
            for j in [n-1,n-3..1] do
                for k in [1..s[j+1]] do
                    u:= f[s[j]]*u;
                od;
                u:= u/(One(LeftActingDomain(L))*Factorial( s[j+1] ) );
            od;
            Add( bash, u );
        od;

        mat:= List( bash, x -> [] );
        sp:= Basis( Subspace( L, [v0] ), [v0] ); 
        for i in [1..Length(bash)] do
            s:= st[i];
            n:= Length(s);
            for j in [i..Length(bash)] do
                u:= bash[j];
                
                for l in [1,3..n-1] do
                    for k in [1..s[l+1]] do
                        u:= e[s[l]]*u;
                    od;
                    u:= u/(One(LeftActingDomain(L))*Factorial( s[l+1] ) );
                od;
                cf:= Coefficients( sp, u );
                if cf = fail then
                   cf:= 0;
                else
                   cf:= cf[1];
                fi;
                mat[i][j]:= cf;
                mat[j][i]:= mat[i][j];
            od;
        od;

        sp:= Basis( Subspace( L, bash ), bash );
        A:= List( BL, h -> Coefficients(sp,h) );

        return A*mat*TransposedMat(A);
           
end;

Jmuseq:= function( R, hw )

        local n, mus, J, B, C, sim, lam, J1, W, ref, Jm, r, r0, ind,
              t, mu;

        n:= Length( CartanMatrix(R) );
        mus:= [hw];
        J:= [ [1..n] ];

        B:= BilinearFormMatNF(R);
        C:= CartanMatrix( R )^-1;
        sim:= SimpleSystemNF(R);
        lam:= hw*C;
        lam:= lam*sim;
        J1:= Filtered( [1..n], i -> lam*B*sim[i] = 0 );
        if Length(J1) > 0 then 
           Add( J, J1 );
        fi;


        W:= WeylGroupAsPermGroup( R );
        ref:= GeneratorsOfGroup( W );

        while Length(J1) > 0 do
            Jm:= J[ Length(J) ];

            for r in [1..n] do
                if not r in Jm then 
                   ind:= Filtered( [1..n], i -> B[r][i] <> 0 );
                   if Length( Intersection( Jm, ind ) ) > 0 then
                      r0:= r;
                      break;
                   fi;
                fi;
            od;

            t:= 1;
            while r in J[t] do t:= t+1; od;
            t:= t-1;
            mu:= ApplyWeylPermToWeight( R, ref[r0], mus[t] );

            Add( mus, mu ); 
            lam:= mu*C;
            J1:= Filtered( Jm, i -> lam*B*sim[i] = 0 );
            if Length(J1) > 0 then 
               Add( J, J1 );
            fi;            
        od;

        return rec( mus:= mus, J:= J );

end;

shortestrep:= function( R, rts, pi )

        # here pi is an element of the Weyl group (given as permutation)
        # rts are some simple roots (as indices in the simple roots of R)
        # generating the reflection subgroup
        # W0; we output the shortest element of W0.pi

        local W, ref, N, v, vi, pos, u;

        W:= WeylGroupAsPermGroup( R );
        ref:= GeneratorsOfGroup( W );

        N:= Length( PositiveRoots(R) );
        v:= pi;
        u:= pi^0;
        while true do
            vi:= v^-1;
            pos:= PositionProperty( rts, i -> i^vi > N );
            if pos = fail then return [u,v]; fi;
            v:= v*ref[rts[pos]];
            u:= ref[rts[pos]]*u;
# note that f : W --> S_n is an anti-homomorphism, so if w[p] denotes the word
# corr to the perm p, then w[v*ref[i]] = w[ref[i]]*w[v], but the latter is
# exactly what we want...        
        od;

end;

Jdecofw0:= function( J, R ) # here J is the J-part of the output of Jmuseq.

        local w0, pi, dec, i, r;

        w0:= LongestWeylWord( R );
        pi:= WeylWordAsPerm( R, w0 );
        dec:= [ ];
        for i in [2..Length(J)] do
            r:= shortestrep( R, J[i], pi );
            Add( dec, r[2] );
            pi:= r[1];
        od;
        Add( dec, pi );

        return dec;
end;


IsInnerAut_simple_real:= function( f )

        # for real forms, so we take a Chev bas rel to max cpt CSA...

        local L, ch, bvecs, hw, BL, A, R, rank, N, nw, Ci, J, Om, rtsNF, rvecs,
              Omvecs, ad, xa, na, conform, CFM, q0, BF, B, m, v, v1, v2, k, s,
              nu, w1, w2, w, p, Phiw, a, vnu, i, j, r, np, u, cf, t, Bnew,
              posr, rts, indsim, word, isinn, one;


        L:= Source(f);

        R:= RootsystemOfCartanSubalgebra( L,
                  MaximallyNonCompactCartanSubalgebra(L) );

        ch:= ChevalleyBasis(R);
        bvecs:= Concatenation( Reversed(ch[1]), ch[3], ch[2] );
        BL:= Basis( L, bvecs );

        A:= List( BL, x -> Coefficients( BL, Image( f, x ) ) );
        A:= TransposedMat(A);
        rank:= Length( CartanMatrix(R) );
        N:= Length(PositiveRoots(R));
        hw:= PositiveRootsAsWeights(R)[N];
        Ci:= CartanMatrix(R)^-1;
        J:= Jmuseq( R, hw );
        Om:= List( J.mus, mu -> mu*Ci );
        rtsNF:= PositiveRootsNF(R);
        rtsNF:= Concatenation( rtsNF, -rtsNF );
        rvecs:= Concatenation( ch[1], ch[2] );
        Omvecs:= List( Om, a -> rvecs[ Position( rtsNF, a ) ] );

# we note that the all elts of Omvecs are root vectors

# the generators of the group

        ad:= List( rvecs, u -> AdjointMatrix( BL, u ) );

        xa:= function( i, t ) # exp (t \ad(x_i) )

            local a, m, b, k;
 
            a:= t*ad[i];
            m:= a^0;
            b:= a; k:= 1;
            while not IsZero(b) do
                m:= m + b;
                b:= b*a; k:= k+1; b:= b/k;
            od;
            return m;
        end;

        na:= function( i )

            local a, j;

            a:= xa( i, -1 );
            if i <= N then
               j:= i+N;
            else
               j:= i-N;
            fi;
            return a*xa( j, 1 )*a;
        end;

        conform:= function( M, u, v )
        # the value of (u,v); M is the matrix of the form
        return Coefficients( BL, u )*M*Coefficients( BL, v );
        end;


        CFM:= conformmat( L,R, ch,  BL );
        q0:= Jdecofw0( J.J, R );
        # now q0[i] is w_{0,i}...

        BF:= BilinearFormMatNF( R );
       
        B:= A;
        word:= [ ];
        for m in [1..Length(J.J)] do
            v:= Coefficients( BL, Omvecs[m] );
            v:= B*v;

            # we are interested in the lowest weight occurring in this...
            k:= Length(v);
            while v[k] = 0 do k:= k-1; od;

            if k <=N then
               s:= N-k+1;
            else # i.e., k > N+rk
               s:= k-rank;
            fi;
# so k is the position in the vector corresponding to the s-th root in the
# normal ordering
            nu:= rtsNF[s]*CartanMatrix(R);

            # find w in W st w(nu) = J.mus[m];

            w1:= ConjugateDominantWeightWithWord( WeylGroup(R), nu )[2];
            w2:= ConjugateDominantWeightWithWord( WeylGroup(R), J.mus[m] )[2];

            w:= Concatenation( w2, Reversed( w1 ) );
            p:= WeylWordAsPerm( R, w );
            w:= PermAsWeylWord( R, p );

            if not ForAll( w, i -> i in J.J[m] ) then
               return rec( isinn:= false );
            fi;

            Phiw:= Filtered([1..N], i -> i^p > N );

            v1:= v;
            a:= v[k];
            vnu:= 0*ShallowCopy(v); vnu[k]:= 1;
            for i in Phiw do
                r:= -2*(rtsNF[s]*BF*rtsNF[i])/(rtsNF[i]*BF*rtsNF[i]);
                np:= na( i )*vnu;
                j:= PositionProperty( np, x -> not IsZero(x) );

                cf:= v1[j]/np[j];
                cf:= cf/(a*(-1)^r);
                #if r <> 1 and not IsZero(cf) then
                #   Print("Need an ",r,"-th root of ",cf,"\n");
                #fi;
                if r = 1 then
                   t:= cf;
                elif r = 2 then
                   t:= Sqroot( cf );
                   v2:= xa( i, -t )*v1;
                   if not IsZero(v2[j]) then
if printstuff then 
   Print("IsInner_aut_real: Approx: ", approx(v2[j]),"\n");
fi;
                      v2:= xa( i, t )*v1;
                      t:= -t;
                      if not IsZero(v2[j]) then Print(v2[j]," Still nonzero!!\n"); Print("Approx: ", approx(v2[j]),"\n");
fi;
                   fi;
                else
                   Print("Higher root needed!!\n");
                fi;
                v1:= xa( i, -t )*v1;
                B:= xa( i, -t )*B;
                if not IsZero(t) then Add( word, ["xl",[i,-t]] ); fi;
            od;

            p:= q0[m]^-1;
            Phiw:= Filtered([1..N], i -> i^p > N );
            for i in Phiw do
                r:= 2*(Om[m]*BF*rtsNF[i])/(rtsNF[i]*BF*rtsNF[i]);
                u:= Coefficients( BL, Omvecs[m] );
                a:= (B*u)[k];
                a:= a*CFM[k][k]/conform( CFM, Omvecs[m], Omvecs[m] );
# note that conform( CFM, vnu, vnu ) = CFM[k][k]...

                u:= na(i)*Coefficients( BL, Omvecs[m] );
                cf:= (B*u)[k];
                cf:= cf*CFM[k][k]/(u*CFM*u);
                cf:= cf/a;

                #if r <> 1 and not IsZero(cf) then
                #   Print("Need an ",r,"-th root of ",cf,"\n");
                #fi;
                if r = 1 then
                   t:= cf;
                elif r = 2 then
                   t:= Sqroot( cf );
                   Bnew:= B*xa( i, -t );
                   v2:= Bnew*u;
                   if not IsZero(v2[k]) then
if printstuff then 
   Print("IsInner_aut_real: Approx: ", approx(v2[k]),"\n");
fi;
                      t:= -t;
                      Bnew:= B*xa( i, -t );
                      v2:= Bnew*u;
                      if not IsZero(v2[k]) then Print(v2[k]," Still nonzero!!\n"); Print("Approx: ", approx(v2[k]),"\n"); fi;
                   fi;
                else
                   Print("Higher root needed!!\n");
                fi;
                B:= B*xa( i, -t );
                if not IsZero(t) then Add( word, ["xr",[i,-t]] ); fi;
            od;
            
            B:= Product( w, i -> na(i) )*B;
            Add( word, ["w",w] );

        od;

        posr:= PositiveRootsNF(R);
        rts:= Reversed( posr );
        Append( rts, List( [1..rank], k -> 0*rts[1] ) );
        Append( rts, -posr );
        indsim:= List( posr{[1..rank]}, x -> Position( rts, x ) );

        if not IsDiagonalMat(B) then return rec( isinn:= false ); fi;

        v:= List( indsim, i -> B[i][i] );
        one:= One( LeftActingDomain(L) );

        isinn:= ForAll( [1..Length(rts)], i -> Product( [1..Length(rts[i])],
                j -> v[j]^rts[i][j] )*one = B[i][i]*one );
        return rec( isinn:= isinn, word:= word, simvals:= v );

end;


check:= function( f, word )


        local L, ch, bvecs, BL, A, ad, xa, na, i, R, rank, N, rvecs;

        L:= Source(f);
        R:= RootsystemOfCartanSubalgebra( L,
                  MaximallyNonCompactCartanSubalgebra(L) );

        ch:= ChevalleyBasis(R);
        bvecs:= Concatenation( Reversed(ch[1]), ch[3], ch[2] );
        BL:= Basis( L, bvecs );
        A:= List( BL, x -> Coefficients( BL, Image( f, x ) ) );
        A:= TransposedMat(A);

        R:= RootSystem(L);
        rank:= Length( CartanMatrix(R) );
        N:= Length(PositiveRoots(R));
        rvecs:= Concatenation( ch[1], ch[2] );

        ad:= List( rvecs, u -> AdjointMatrix( BL, u ) );

        xa:= function( i, t ) # exp (t \ad(x_i) )

            local a, m, b, k;
 
            a:= t*ad[i];
            m:= a^0;
            b:= a; k:= 1;
            while not IsZero(b) do
                m:= m + b;
                b:= b*a; k:= k+1; b:= b/k;
            od;
            return m;
        end;

        na:= function( i )

            local a, j;

            a:= xa( i, -1 );
            if i <= N then
               j:= i+N;
            else
               j:= i-N;
            fi;
            return a*xa( j, 1 )*a;
        end;

        for i in [1..Length(word)] do
            if word[i][1] = "xl" then
               A:= xa(word[i][2][1],word[i][2][2])*A;
            elif word[i][1] = "xr" then
               A:= A*xa(word[i][2][1],word[i][2][2]);
            elif word[i][1] = "w" then
               A:= Product( word[i][2], j -> na(j) )*A;
            fi;
        od;
        return List( [1..Length(A)], i -> A[i][i] );
end;

# In the next function C is an lxl Cartan matrix, u is a vector with l
# coordinates. We return [ t1,...,tl] such that
# t1^C(j,1)....tl^C(j,l) = uj

diagonalaut:= function( C, u ) 

        local rk, r, P, H, v, j, t, q;

        rk:= Length( C );

        r:= HermiteNormalFormIntegerMatTransforms( C );
        P:= r.rowtrans;
        H:= r.normal;

        v:= [ ];
        for j in [1..rk] do
            v[j]:= Product( [1..rk], i -> u[i]^P[j,i] );
        od;

        t:= [ ];
        for j in [rk,rk-1..1] do
            q:= v[j]*Product( [j+1..rk], i -> t[i]^-H[j,i] );
            q:= q*One( SqrtField );
            t[j]:= nRoot( H[j][j], q );
        od;

        return t;

end;

inneraut_asword:= function( f )


        local left, right, diag, L, R, ch, N, bvecs, r, i, j, k, word, t, w;

        left:= [ ];
        right:= [ ];
        diag:= [ ];

        L:= Source(f);

        R:= RootsystemOfCartanSubalgebra( L,
                  MaximallyNonCompactCartanSubalgebra(L) );
        N:= Length( PositiveRootsNF(R) );

        r:= IsInnerAut_simple_real( f );

        if not r.isinn then
           return fail;
        fi;

        word:= r.word;

        for i in [1..Length(word)] do
            if word[i][1] = "xl" then
               Add( left, [ word[i][2][1], -word[i][2][2] ] );
            elif word[i][1] = "xr" then
               Add( right, [ word[i][2][1], -word[i][2][2] ] );
            elif word[i][1] = "w" then
               w:= Reversed( word[i][2] );
               for j in w do
                   if j <= N then
                      k:= j+N;
                   else
                      k:= j-N;
                   fi;
                   Add( left, [ j, 1 ] );
                   Add( left, [ k, -1 ] );
                   Add( left, [ j, 1 ] );
               od;
            fi;
        od;

        t:= diagonalaut( CartanMatrix(R), r.simvals );
        for i in [1..Length(t)] do
            Add( diag, [ i, t[i] ] );
            Add( diag, [ i+N, -t[i]^-1 ] );
            Add( diag, [ i, t[i] ] );
            Add( diag, [ i, -1 ] );
            Add( diag, [ i+N, 1 ] );
            Add( diag, [ i, -1 ] );
        od;


        return Concatenation( left, diag, Reversed(right) );
end;


checkval:= function(f, word, x )

        # x in Source(f) we check that we get the same by applying f and word

        local L, R, ch, rvecs, ad, xa, cf, w, i;

        L:= Source(f);

        R:= RootsystemOfCartanSubalgebra( L,
                  MaximallyNonCompactCartanSubalgebra(L) );

        ch:= ChevalleyBasis(R);
        rvecs:= Concatenation( ch[1], ch[2] );

        ad:= List( rvecs, u -> AdjointMatrix( Basis(L), u ) );

        xa:= function( i, t ) # exp (t \ad(x_i) )

            local a, m, b, k;
 
            a:= t*ad[i];
            m:= a^0;
            b:= a; k:= 1;
            while not IsZero(b) do
                m:= m + b;
                b:= b*a; k:= k+1; b:= b/k;
            od;
            return m;
        end;

        cf:= Coefficients( Basis(L), x );

        w:= Reversed( word );
        for i in [1..Length(w)] do
            cf:= xa( w[i][1], w[i][2] )*cf;
        od;

        #cf:= SqrtFieldMakeRational(cf);

        return cf*Basis(L) = Image( f, x );
end;


conjmat:= function( f )

        # f is an inner automorphism of a *simple* matrix Lie algebra.
        # we obtain a matrix g such that gxg^-1 = f(x) for x in the Lie alg.

        local word, L, R, ch, rv, xa, xa0;

        word:=inneraut_asword( f );
        L:= Source( f );

        R:= RootsystemOfCartanSubalgebra( L,
                  MaximallyNonCompactCartanSubalgebra(L) );
        ch:= ChevalleyBasis(R);
        rv:= List( Concatenation( ch[1], ch[2] ), x -> x![1] );

        xa:= function( i, t ) # exp (t \x_i )

            local a, m, b, k;
 
            a:= t*rv[i];
            m:= a^0;
            b:= a; k:= 1;
            while not IsZero(b) do
                m:= m + b;
                b:= b*a; k:= k+1; b:= b/k;
            od;
            return m;
        end;

        xa0:= function( i, t )

            local a, n;

            a:= t*rv[i];
            n:= Length(a);
            return Sum( [0..n], j -> a^j/Factorial(j) );
        end;

        return Product( List( word, w -> xa( w[1], w[2] ) ) );
end;

normalizecoc:= function( r, c )

        # r:  the output of H1connected
        # c: a cocycle in the connected group corr to r.L
        # we find g in the identity component G^0 such that
        # g^-1 * c * bar{g} is equal to one of the standard cocycles
        # in r.cocyc. 

        local L, F, n, jd, s, u, u1, lgu, u2, i, ex, CM, M, V, C, Tp, H0,
              DL, f, dd, mts, f0, pieces, t, R, tp, type, perm, c1, c2,
              b1, b2, en, H, QQ, r0, Q, A, U, diag, v, m, inds, tau, a, pos,
              k, is0, SS0, j, ts, gg, g, l, ub, b, g1, s1, z, s0, mm, BM, aa,
              sol, pi, f1, C0, list, cc1, cc2;

        L:= r.L;
        F:= LeftActingDomain( L );
        n:= Length( Basis(L)[1]![1] );

        jd:= JordanDecomposition( c );
        s:= jd[1];
        u:= s^-1*c;

        u1:= u-u^0;
        lgu:= Zero(F)*u;
        u2:= u1;
        i:= 1;
        while not IsZero(u2) do
            lgu:= lgu+(-1)^(i-1)*u2/i;
            u2:= u2*u1; i:= i+1;
        od;

        lgu:= lgu/2;
        ex:= lgu^0;
        u2:= lgu; i:= 1;
        while not IsZero(u2) do
            ex:= ex + u2/Factorial(i);
            u2:= u2*lgu; i:= i+1;
        od;

        # now we have ex^-1*c*\bar{ex} = s.
        # so we continue with s

        mm:= List( Basis(L), x -> x![1] );
        M:= VectorSpace( SqrtField, mm );
        BM:= Basis( M, mm );
        aa:= List( mm, x -> Coefficients( BM, s*x*s^-1 ) );
        for i in [1..Length(aa)] do aa[i][i]:= aa[i][i]-1; od; 
        sol:= NullspaceMat( aa );
        C:= Subalgebra( L, List( sol, x -> LieObject( x*mm ) ) );

        QQ:= r.QQ;

        Tp:= CartanSubalgebra( C );
        if not ForAll( Basis(Tp), x -> LieObject(QQ*ComplexConjugate(x)*QQ^-1)
                in Tp ) then
           SetRealStructure( C, function(x)
                     return LieObject(QQ*ComplexConjugate(x)*QQ^-1); end );
           Tp:= sigmaCSA( C );
        fi;

        C0:= LieCentralizer( L, Tp );
        H0:= CartanSubalgebra( C0 );
        if not ForAll( Basis(H0), x -> LieObject(QQ*ComplexConjugate(x)*QQ^-1)
                in H0 ) then
           SetRealStructure( C0, function(x)
                     return LieObject(QQ*ComplexConjugate(x)*QQ^-1); end );
           H0:= sigmaCSA( C0 );
        fi;

        # we find the conjugation on the torus corr to H0...
        r0:= toruspar_gen( F, List( Basis(H0), x -> x![1] ) );
        Q:= r0.Q;
        A:= r0.C^-1*QQ*ComplexConjugate( r0.C );
        U:= A*Product( r0.mats )*A^-1;
        diag:= List( [1..Length(U)], i -> U[i][i] );
        v:= List( Q, s -> Product( List( [1..Length(s)], i -> diag[i]^s[i])));

        m:= Dimension(H0);
        inds:= List( r0.indets, x -> ExtRepPolynomialRatFun(x)[1][1] );
        tau:= NullMat(m,m);
        for i in [1..m] do
            a:= ExtRepNumeratorRatFun( v[i] );
            for k in [1,3..Length(a[1])-1] do
                pos:= Position( inds, a[1][k] );
                tau[i][pos]:= tau[i][pos]+a[1][k+1];
            od;
            a:= ExtRepDenominatorRatFun( v[i] );
            for k in [1,3..Length(a[1])-1] do
                pos:= Position( inds, a[1][k] );
                tau[i][pos]:= tau[i][pos]-a[1][k+1];
            od;
        od;

        is0:= isomTS( tau );

        # compute the subtori on the "S"-side...
        SS0:= [ ];
        for i in [1..Length(r0.indets)] do
            v:= List( r0.indets, x -> 1 );
            v[i]:= r0.indets[i];
            v:= is0.StoT( v );
            M:= r0.C*Product( r0.mats )*r0.C^-1;
            for i in [1..Length(M)] do
                for j in [1..Length(M)] do
                    M[i][j]:= Value( M[i][j], r0.indets, v );
                od;
            od;
            Add( SS0, M );
        od;

        t:= valpars( r0, s );
        ts:= is0.TtoS( t );

        gg:= [ ];

        for i in [is0.flen+1..is0.flen+is0.elen] do
            u:= ts[i];
            if not IsOne(u) then
               if u = -u^0 then
                  v:= E(4)*u^0;
               else
                  ub:= ComplexConjugate(u);
                  a:= (u+ub)/2;
                  b:= (u-ub)/(2*E(4));
                  v:= b+(-a+1)*E(4);
               fi;
               g:= List( SS0[i], ShallowCopy );
               for k in [1..Length(g)] do
                   for l in [1..Length(g)] do
                       g[k][l]:= value( g[k][l], [ r0.indets[i] ], [v] );
                   od;
               od;
               Add( gg, g );
               ts[i]:=1;
            fi;
        od;

        for i in [is0.flen+is0.elen+1..is0.flen+is0.elen+is0.glen] do
            v:= ts[i];
            g:= List( SS0[i], ShallowCopy );
            for k in [1..Length(g)] do
                for l in [1..Length(g)] do
                    g[k][l]:= value( g[k][l], [ r0.indets[i] ], [v] );
                od;
            od;
            Add( gg, g );
            ts[i]:=1; ts[is0.glen+i]:=1;
        od;

if printstuff then
   g:= Product(gg);
   t:= valpars( r0, g^-1*s*ComplexConjugate(g) );
   Print("normalizecoc: check ts...",is0.TtoS( t ),"\n");;
fi;


        if ForAll( ts, IsOne ) then
           if Length(gg)=0 then
              return IdentityMat( n, F );
           else
              return Product(gg);
           fi;
        fi;

        DL:= LieDerivedSubalgebra( L );
        U:= Intersection( DL, H0 );
        H0:= maxcpttorus( DL, U );

        f:= autom( DL, H0 );

        H:= MaximallyCompactCartanSubalgebra( DL );

if printstuff then 
    Print("normalizecoc: check on compact CSA... ", ForAll( Basis(H), x -> Image(f,x) in H0 ),"\n");
fi;

        dd:= DirectSumDecomposition( DL );

        # first we see if f permutes certain summands, if yes we
        # construct an automorphism performing the inverse permutation,
        # and compose..

        list:= List( dd, U -> PositionProperty( dd, M -> Image(f,Basis(U)[1])
                                in M ) );
        perm:= PermList( list );
        if perm <> perm^0 then

           R:= List(dd, U -> RootsystemOfCartanSubalgebra(U,Intersection(U,H)));
           c1:= List( R, CanonicalGenerators );
           c2:= [ ];
           for i in [1..Length(list)] do
               k:= list[i];
               c2[k]:= c1[i];
           od;

           cc1:= [ Concatenation( List( c1, u -> u[1] ) ),
                   Concatenation( List( c1, u -> u[2] ) ),
                   Concatenation( List( c1, u -> u[3] ) ) ];

           cc2:= [ Concatenation( List( c2, u -> u[1] ) ),
                   Concatenation( List( c2, u -> u[2] ) ),
                   Concatenation( List( c2, u -> u[3] ) ) ];
           b1:= Concatenation( SLAfcts.canbas( DL, cc1 ) );
           b2:= Concatenation( SLAfcts.canbas( DL, cc2 ) );
           f:= AlgebraHomomorphismByImagesNC(DL,DL,b1,b2)*f;
        fi;

        pieces:= [ ];
        for i in [1..Length(dd)] do
            f0:= AlgebraHomomorphismByImagesNC( dd[i], dd[i], Basis(dd[i]),
                        List( Basis(dd[i]), x -> Image( f, x ) ) );
            t:= IsInnerAut_simple_real(f0);
if printstuff then
   Print("normalizecoc: f0 is inner... ",t.isinn,"\n");
fi;

            if t.isinn then
               Add( pieces, f0 );
            else
               R:= RootsystemOfCartanSubalgebra(dd[i], Intersection( dd[i], H ) );
               tp:= CartanType( CartanMatrix(R) );

               type:= tp.types[1][1];
               en:= tp.enumeration[1];
               if type = "A" then
                  perm:= [ MappingPermListList( en, Reversed(en) ) ];

               elif type ="D" then
                  if Length( en ) >= 5 then
                     perm:= [ MappingPermListList( en,
                                  Permuted(en,(Length(en)-1,Length(en))) ) ];
                  else
                     perm:= [ MappingPermListList( en, Permuted(en,(3,4)) ),
                              MappingPermListList( en, Permuted(en,(1,3)) ),
                              MappingPermListList( en, Permuted(en,(1,4)) ),
                              MappingPermListList( en, Permuted(en,(1,3,4)) ),
                              MappingPermListList( en, Permuted(en,(1,4,3)) )];
                  fi;
               else
                  perm:= [ MappingPermListList( en,
                                Permuted(en,(1,6)*(3,5)) ) ];
               fi;
if printstuff then
   Print("normalizecoc: perms... ",perm,"\n");
fi;
               c1:= CanonicalGenerators( R );
               b1:= Concatenation( SLAfcts.canbas( dd[i], c1 ) );
               for pi in perm do 
                   c2:= List( c1, x -> Permuted( x, pi ) );
                   b2:= Concatenation( SLAfcts.canbas( dd[i], c2 ) );
                   f1:= AlgebraHomomorphismByImagesNC( dd[i], dd[i],
                                                           b1, b2 )*f0;
                   t:= IsInnerAut_simple_real(f1);
                   if t.isinn then
                      Add( pieces, f1 );
                      break;
                   fi;
               od;

            fi;
        od;

        mts:= List( pieces, h -> conjmat(h) );

        g:= Product( mts );

if printstuff then
   Print("normalizecoc: check on g....",ForAll( Basis(H), x ->  LieObject( g*x![1]*g^-1) in H0 ),"\n");
fi;

        z:= g^-1*QQ*ComplexConjugate(g)*QQ^-1;
        if not z=z^0 then
if printstuff then 
   Print("normalizecoc: z is cocycle... ", z*QQ*ComplexConjugate(z)*QQ^-1 = z^0, " (should be true)\n" );
fi;
           u:= normalize_toruscoc( r, z );
           g:= g*u;
        fi;

        g1:= Product(gg);
        s1:= g1^-1*s*QQ*ComplexConjugate(g1)*QQ^-1;
        s1:= g^-1*s1*g;

# now s1 should lie in the standard max cpt torus...
if printstuff then 
   t:= valpars( r.torusdata, s1 ); 
   Print("normalizecoc: check on s1 in torus.... ",t,"(shoud not be fail) \n");
fi;

        u:= normalize_toruscoc( r, s1 );

        return ex*g1*g*u;

end;


SCHweights:= function( C )

        # C: Cartan matrix, we compute a list of highest weights such that
        # the group constructed with the corresponding module is
        # simply connected

        local tp, wts, n, i, type, en, w, rk;

        tp:= CartanType( C );
        wts:= [ ];
        n:= Length(C);
        for i in [1..Length(tp.types)] do
            type:= tp.types[i][1]; rk:= tp.types[i][2];
            en:= tp.enumeration[i];
            if type = "A" or type = "C" then
               w:= 0*[1..n];
               w[en[1]]:= 1;
               Add( wts, w );
            elif type = "B" then
               w:= 0*[1..n];
               w[en[rk]]:= 1;
               Add( wts, w );
            elif type = "D" then
               w:= 0*[1..n];
               w[en[rk]]:= 1;
               Add( wts, w );
               if IsEvenInt(rk) then
                  w:= 0*[1..n];
                  w[en[rk-1]]:= 1;
                  Add( wts, w );
               fi;
            elif type = "E" and rk=6 then
               w:= 0*[1..n];
               w[en[1]]:= 1;
               Add( wts, w );
            elif type = "E" and rk in [7,8] then
               w:= 0*[1..n];
               w[en[rk]]:= 1;
               Add( wts, w );
            elif type = "F" and rk=4 then
               w:= 0*[1..n];
               w[en[rk]]:= 1;
               Add( wts, w );
            else
               w:= 0*[1..n];
               w[en[1]]:= 1;
               Add( wts, w );
            fi;

        od;

        return wts;
end;

#PRU:= function( k ) # primitive root of unity in SqrtField...
#
#        local a;
#
#        if k=1 then
#           return One( SqrtField );
#        elif k=2 then
#           return -One(SqrtField);
#        elif k=3 then
#           return (-1+Sqroot(-3))/2;
#        elif k=4 then
#           return E(4)*One( SqrtField );
#        elif k=5 then
#           a:= Sqroot(5);
#           return ( Sqroot( 2*a-10 )-a-1)/4;
#        elif k=6 then
#           return (1-Sqroot(-3))/2; 
#        fi;
#
#end;

ScCentreC:= function( C )

        # here C is a Cartan matrix
        # we return the centre of the simply connected group
        # the output is a list of lists [t1,...,tn] representing
        # the element h_{alpha_1}(t1)....h_{alpha_n}(tn).

        local H, dd, n, i, j, k, dd0, d, v, exp, z;

        H:= HermiteNormalFormIntegerMat(C);
        dd:= [ [] ];
        n:= Length(H);
        k:= n;
        while k >= 1 do
              dd0:= [ ];
              for i in [1..Length(dd)] do
                  d:= dd[i];
                  v:= One( SqrtField );
                  for j in [k+1..n] do
                      v:= v*(d[j-k]^-H[k][j]);
                  od; 
                  exp:= H[k][k];
                  if exp > 1 then
                     v:= nRoot( exp, v );
                  fi;
                  z:= PRU( exp );
                  for j in [0..exp-1] do
                      Add( dd0, Concatenation( [ z^j*v ], d ) );
                  od;
              od;
              dd:= dd0;
              k:= k-1;
        od;

        return dd;
end;

ScCentre:= function( L, QQ )

        # here L is a semisimple matrix Lie algebra with conjugation
        # A |---> QQ*ComplexConjugate(A)*QQ^-1.
        # We return the centre of the corresponding simply connected group,
        # as in the previous function, with conjugation.

        local sig, H, R, C, cc, rank, ww, V, e, f, xa, wa, wa1, ha, wabar,
              wa1bar, habar, elms, conjs, bas, BL, T, i, j, K, N, LtoK,
              RK, mV, mat, ebar, fbar, U, BU, imsig, basU, UtoL;

        sig:= function( A ) return QQ*LieObject(ComplexConjugate(A))*QQ^-1; end;

        H:= MaximallyNonCompactCartanSubalgebra( L );
        R:= RootsystemOfCartanSubalgebra(L,H);
        C:= CartanMatrix( R );
        cc:= ScCentreC( C );

        rank:= Length( C );

        # we make a rational copy of the Lie algebra wrt Chevalley basis...
        bas:= Concatenation( ChevalleyBasis(R) );
        BL:= Basis( L, bas );
        T:= ShallowCopy( StructureConstantsTable(BL) );
        for i in [1..Dimension(L)] do
            T[i]:= ShallowCopy( T[i] );
            for j in [1..Dimension(L)] do
                T[i][j]:= ShallowCopy( T[i][j] );
                T[i][j][2]:= List( T[i][j][2], SqrtFieldMakeRational );
            od;
        od;
        T[Length(T)]:= 0;
        K:= LieAlgebraByStructureConstants( Rationals, T );
        N:= Length( PositiveRoots(R) );
        SetCartanSubalgebra( K, Basis(K){[2*N+1..2*N+rank]} );
        LtoK:= function( u )
             return List( Coefficients( BL, u ), SqrtFieldMakeRational )*
                                                  Basis(K);
        end;  

        RK:= Objectify( NewType( NewFamily( "RootSystemFam", IsObject ),
                     IsAttributeStoringRep and IsRootSystemFromLieAlgebra ), 
                     rec() );
        SetCanonicalGenerators( RK, List(CanonicalGenerators(R),
                                                 x->List(x,LtoK)));
        SetUnderlyingLieAlgebra( RK, K );
        SetPositiveRootVectors( RK, List(PositiveRootVectors(R),LtoK));
        SetNegativeRootVectors( RK, List(NegativeRootVectors(R),LtoK));
        SetCartanMatrix( RK,  CartanMatrix(R) );
        SetPositiveRoots( RK, PositiveRoots(R));
        SetNegativeRoots( RK, NegativeRoots(R));
        SetSimpleSystem( RK, SimpleSystem(R));
        SetRootSystem( K, RK);
        SetChevalleyBasis(K,List(ChevalleyBasis(R),x->List(x,LtoK)));

        ww:= SCHweights( C );
        V:= DirectSumOfAlgebraModules( List( ww, w ->
                HighestWeightModule( K, w ) ) );
        mV:= List( Basis(K), x -> MatrixOfAction( Basis(V), x ) );
        mV:= mV*One( SqrtField );

        mat:= function( u ) # u in L...
             return Coefficients( BL, u )*mV;
        end;

        e:= List( CanonicalGenerators(R)[1], mat);
        f:= List( CanonicalGenerators(R)[2], mat);

        xa:= function( u, t ) # exp( t*u )


            local a, m, b, k;

            a:= t*u;
            m:= a^0;
            b:= a; k:= 1;
            while not IsZero(b) do
                m:= m + b;
                b:= b*a; k:= k+1; b:= b/k;
            od;
            return m;
        end;


        wa:= function( i, t ) # w_{alpha_i}(t)

             local u;

             u:= xa( e[i], t  );
             return u*xa(f[i],-t^-1)*u;
        end;

        wa1:= List( [1..rank], i -> wa( i, 1 )^-1 );

        ha:= function( i, t ) # h_{alpha_i}(t)

             return wa( i, t )*wa1[i];
        end;

        ebar:= List( CanonicalGenerators(R)[1], x -> mat(sig(x)) );
        fbar:= List( CanonicalGenerators(R)[2], x -> mat(sig(x)) );

        wabar:= function( i, t ) # conjugate w_{alpha_i}(t)

             local u;

             u:= xa( ebar[i], ComplexConjugate(t) );
             return u*xa(fbar[i],ComplexConjugate(-t^-1))*u;
        end;

        wa1bar:= List( [1..rank], i -> wabar( i, 1 )^-1 );

        habar:= function( i, t ) # h_{alpha_i}(t)

             return wabar( i, t )*wa1bar[i];
        end;

        elms:= List( cc, c -> Product( [1..rank], i -> ha( i, c[i] ) ) );
        conjs:= List( cc, c -> Product( [1..rank], i -> habar( i, c[i] ) ) );

        U:= LieAlgebra( SqrtField, mV );

        basU:= Basis( U, List( mV, LieObject ) );
        UtoL:= function( u )
            local a;
            if not IsLieObject(u) then
               a:= LieObject(u);
            else
               a:= u;
            fi;
            return Coefficients( basU, u )*bas;
        end;


        BU:= BasisNC( U, List( mV, LieObject ) );
        imsig:= List( bas, x -> Coefficients( BL, sig(x) ) );

        SetRealStructure( U,
         function( v )
             return (List( Coefficients(BU,v), ComplexConjugate )*imsig)*BU;
         end );

        T:= maxcptCSA( U );

        return rec( cc:= cc, conj:= List( conjs, x -> Position( elms, x ) ),
                    scelms:= elms, maxcpt:= List( Basis(T), x -> x![1]),
                    U:= U, UtoL:= UtoL);
end;

H1disconnected_torus:= function( mats, tab, gam, pinv )

        # This is the same as H1disconnected, except that we assume that the
        # identity component of the group is a torus. This makes life
        # much easier...

        # everything over SqrtField

        # here:

        # mats - matrices forming a basis of the Lie algebra
        # tab - a multiplication table of the component group,
        #       so mat[i][j]=k means g_i*g_j = g_k 1 <= i,j,k <= n
        # gam - is a function giving the conjugation on the component group,
        #       so the conjugate of i is gam(i)
        # pinv - is a list of elements of the group giving the inverse
        #        images of the elements of the component group, so
        #        for 1 <= i <=n we have that pinv[i] is an inverse image
        #        of i. 
        
        local L, D, sig, TD, R, sim, cg1, cb1, cg2, cb2, map, dd, pieces,
              i, f0, t, c1, inds, tp, type, en, perm, b1, c2, b2, mts, a0,
              H1C, g, h, a, hp, fp, sc, z2, h2, u1, u2, u, classes, cls,
              j, imh2, xa, wa, wa1, ha, e, f, centraltorus, rC, imC, found,
              c, s1, pos, Q, A, U, diag, v, m, tau, k, is, v1, l, M, rT, b, h1,
              q, n0, one, inv, prod, Ctau, act, c0, HH1, orbs, M0, wd,
              dU, Rd, ch, rv;

        n0:= Length( tab );
        one:= Position( tab, [1..n0] );

        inv:= function( i )
               return Position( tab[i], one );
        end;

        prod:= function( i, j )
              return tab[i][j];
        end;

        L:= LieAlgebra( SqrtField, mats );

        H1C:= H1finite( tab, gam );
        HH1:= [ ];

        rC:=  toruspar_gen( SqrtField, List( Basis(L), x -> x![1] ) );

        for q in H1C do

            g:= pinv[q];
            h:= g*ComplexConjugate(g);  # later use more general conjugation...

if printstuff then 
   Print("H1disconnected_torus: g is cocycle... ", h=h^0,"\n");
fi;
            # we ned to see whether h in B^2 T,  T=G^0, where we use the
            # conjugation twisted by g...

            # First we compute the conjugation of T, and an
            # isomorphism to a torus in "normal form", that is a
            # product of standard tori.

            Q:= rC.Q;
            A:= rC.C^-1*g*ComplexConjugate( rC.C );
            U:= A*Product( rC.mats )*A^-1;
            diag:= List( [1..Length(U)], i -> U[i][i] );
            v:= List( Q, s -> Product( List( [1..Length(s)], i ->
                                                         diag[i]^s[i])));
            m:= Dimension(L);
            inds:= List( rC.indets, x->ExtRepPolynomialRatFun(x)[1][1] );
            tau:= NullMat(m,m);
            for l in [1..m] do
                u:= ExtRepNumeratorRatFun( v[l] );
                for k in [1,3..Length(u[1])-1] do
                    pos:= Position( inds, u[1][k] );
                    tau[l][pos]:= tau[l][pos]+u[1][k+1];
                od;
                u:= ExtRepDenominatorRatFun( v[l] );
                for k in [1,3..Length(u[1])-1] do
                    pos:= Position( inds, u[1][k] );
                    tau[l][pos]:= tau[l][pos]-u[1][k+1];
                od;
            od;

            # the element lies in B^2 T iff the parameters corresponding
            # to the subtori isomorphic to E (consisting of u\in C^*
            # with upgam u = \bar{u}) are positive. If such a parameter
            # is not positive then we set found to false again.
            # Otherwise we find the element M of the group such that
            # the elt we deal with is equal to M * upgam M. 

            
            is:= isomTS( tau );
            v:= valpars( rC, h );
            v1:= is.TtoS(v);
            v:= [ ];
            found:= true;
            for k in [1..is.flen] do
                v[k]:= Sqroot(v1[k]);
            od;
            for k in [is.flen+1..is.flen+is.elen] do
                if ispositive(v1[k]) = false then
                   found:= false;
                   break;
                elif ispositive(v1[k]) = true then
                   v[k]:= Sqroot(v1[k]);
                else
                   Print("ERROR!!, not clear if ",v1[k]," is positive."); 
                fi;
            od;
            if found then
               # finish v
               for k in [is.flen+is.elen+1..is.flen+is.elen+is.glen] do
                   v[k]:= v1[k];
                   v[k+is.glen]:= One(SqrtField);
               od;

               v:= is.StoT(v);
               M:= rC.C*Product( rC.mats )*rC.C^-1;
               for k in [1..Length(M)] do
                   for l in [1..Length(M)] do
                       M[k][l]:= Value( M[k][l], rC.indets, v );
                   od;
               od;

               u:= M^-1*g;

if printstuff then 
   Print("H1disconected_torus: u is a cocycle... ",u*ComplexConjugate(u)=u^0,"(should be true) \n");
fi;
# so u is a cocycle...


               h1:= H1connected( mats, u );

               # now we determine C^\tau...
               # q is the cocycle by which we twist in C...

               Ctau:= Filtered( [1..n0], k ->
                            prod( q, prod( gam(k), inv(q) ) ) =k ); 

               # now we compute the action of Ctau on h1...

               act:= List( [1..Length(h1.cocyc)], x -> [ ] );

               # act[k] contains the images of h1 under the elts
               # of Ctau...

               for k in [1..Length(act)] do
                   for l in [1..Length(Ctau)] do
                       b:= pinv[ Ctau[l] ];

                       c0:= b^-1*h1.cocyc[k]*(u*ComplexConjugate(b)*u^-1);
if printstuff then
   Print("H1disconnected_torus: c0 is cocycle... ",c0*(u*ComplexConjugate(c0)*u^-1)=c0^0,"(should be true)\n");
fi;
                       g:= normalize_toruscoc( h1, c0 );
                       c0:= g^-1*c0*(u*ComplexConjugate(g)*u^-1);
                       act[k][l]:= Position( h1.cocyc, c0 );
                   od;
               od;

               orbs:= Set( List( act, Set ) );
               Append( HH1, List( orbs, x -> h1.cocyc[x[1]]*u ) );

            fi;

        od;

        return HH1;
end;


neutralize2cocreductive:= function( D, T, R, rC, g )

        # D: the derived algebra (so the semisimple part),
        # T: the centre
        # R: root sys of D wrt max cpt Cartan
        # rC: the output of toruspar_gen with input T
        # g: the element so that we need to trivialize [h,f] where
        # h = g (gam*g), f = inn(g)\circ \sigma. 

        local sig, h, Q, A, U, diag, v, m, inds, tau, l, u, k, pos, is,
              v1, found, M, cg1, cb1, cg2, cb2, map, dd, cgs1, cgs2,
              mapi, pieces, i, t, tp, type, en, perm, pi, c1, b1,
              c2, f1, mts, a, hp, sc, dU, z2, b2, u1, u2, classes, cls,
              j, h2, imh2, imC, c, s1, centraltorus, rT, b, M0, f0, d0, wd, Rd,
              ch, rv, e, f, xa, wa, wa1, ha;

        if R = fail then
           e:= [ ]; f:= [ ];
        else
           e:= List( CanonicalGenerators(R)[1], x -> x![1] );
           f:= List( CanonicalGenerators(R)[2], x -> x![1] );
        fi;

        xa:= function( u, t ) # exp( t*u )


            local a, m, b, k;

            a:= t*u;
            m:= a^0;
            b:= a; k:= 1;
            while not IsZero(b) do
                m:= m + b;
                b:= b*a; k:= k+1; b:= b/k;
            od;
            return m;
        end;

        wa:= function( i, t ) # w_{alpha_i}(t)

             local u;

             u:= xa( e[i], t  );
             return u*xa(f[i],-t^-1)*u;
        end;

        wa1:= List( [1..Length(e)], i -> wa( i, 1 )^-1 );

        ha:= function( i, t ) # h_{alpha_i}(t)

             return wa( i, t )*wa1[i];
        end;


        sig:=function( v ) return LieObject(ComplexConjugate(v)); end;

        h:= g*ComplexConjugate(g); 

        if Dimension(D)=0 then

           # we need to see whether h in B^2 T,  T=G^0, where we use the
           # conjugation twisted by g...

           # First we compute the conjugation of T, and an
           # isomorphism to a torus in "normal form", that is a
           # product of standard tori.

           Q:= rC.Q;
           A:= rC.C^-1*g*ComplexConjugate( rC.C );
           U:= A*Product( rC.mats )*A^-1;
           diag:= List( [1..Length(U)], i -> U[i][i] );
           v:= List( Q, s -> Product( List( [1..Length(s)], i ->
                                                         diag[i]^s[i])));
           m:= Dimension(T);
           inds:= List( rC.indets, x->ExtRepPolynomialRatFun(x)[1][1] );
           tau:= NullMat(m,m);
           for l in [1..m] do
               u:= ExtRepNumeratorRatFun( v[l] );
               for k in [1,3..Length(u[1])-1] do
                   pos:= Position( inds, u[1][k] );
                   tau[l][pos]:= tau[l][pos]+u[1][k+1];
               od;
               u:= ExtRepDenominatorRatFun( v[l] );
               for k in [1,3..Length(u[1])-1] do
                   pos:= Position( inds, u[1][k] );
                   tau[l][pos]:= tau[l][pos]-u[1][k+1];
               od;
           od;

           # the element lies in B^2 T iff the parameters corresponding
           # to the subtori isomorphic to E (consisting of u\in C^*
           # with upgam u = \bar{u}) are positive. If such a parameter
           # is not positive then we return fail.
           # Otherwise we find the element M of the group such that
           # the elt we deal with is equal to M * upgam M. 

            
           is:= isomTS( tau );
           v:= valpars( rC, h );
           v1:= is.TtoS(v);
           v:= [ ];
           for k in [1..is.flen] do
               v[k]:= Sqroot(v1[k]);
           od;
           for k in [is.flen+1..is.flen+is.elen] do
               if ispositive(v1[k]) = false then
                  return fail;
               elif ispositive(v1[k]) = true then
                  v[k]:= Sqroot(v1[k]);
               else
                  Print("ERROR!!, not clear if ",v1[k]," is positive."); 
               fi;
           od;
           for k in [is.flen+is.elen+1..is.flen+is.elen+is.glen] do
               v[k]:= v1[k];
               v[k+is.glen]:= One(SqrtField);
           od;

           v:= is.StoT(v);
           M:= rC.C*Product( rC.mats )*rC.C^-1;
           for k in [1..Length(M)] do
               for l in [1..Length(M)] do
                   M[k][l]:= Value( M[k][l], rC.indets, v );
               od;
           od;

           return M^-1;
        
        else # dim D > 0

           centraltorus:= Dimension(T) > 0; 
           cg1:= CanonicalGenerators(R);
           cb1:= R!.canbas1;

           cg2:= List( cg1, x -> List( x, u -> g*sig(u)*g^-1 ) );         
           cb2:= Concatenation( SLAfcts.canbas( D, cg2 ) );
           map:= AlgebraHomomorphismByImagesNC( D, D, cb2, cb1 );

      # here map^-1 : x ----> g sig(x) g^-1
      # we find a in the identity component of the group such that
      # map^-1(x) = a x a^-1, apart from possibly a permutation of the simple
      # roots, so if r(x) = a x a^-1 and f(x) = g sig(x) g^-1 (=map^-1)
      # then r ° f is an outer automorphism permuting the simple root vectors.

      # We have that Delta(q) = ( h, f ) where f = map^-1, lies in Z^2 G^circ
      # We the modify it to get ( h', f' ) with h' = a f(a) h (=hp),
      # f' = inn(a)°f. 

           dd:= DirectSumDecomposition( D );

           if not ForAll(dd,A-> ForAll(Basis(A),a-> Image(map,a) in A)) then
              # compose with an outer automorphism so that the direct sum
              # decomposition is respected
              cgs1:=[ ]; cgs2:= [ ];
              for i in [1..Length(dd)] do
                  j:= PositionProperty( dd,
                                     A->Image(map,Basis(dd[i])[1]) in A);
                  c1:= List( cg1, u -> Filtered( u, x -> x in dd[i] ) );
                  cgs2[j]:= c1;
                  cgs1[i]:= c1;
              od;

              cg1:= [ Concatenation( List( cgs1, k -> k[1] ) ),
                      Concatenation( List( cgs1, k -> k[2] ) ),
                      Concatenation( List( cgs1, k -> k[3] ) )];

              cg2:= [ Concatenation( List( cgs2, k -> k[1] ) ),
                       Concatenation( List( cgs2, k -> k[2] ) ),
                       Concatenation( List( cgs2, k -> k[3] ) )];
              cb1:= Concatenation( SLAfcts.canbas( D, cg1 ) );
              cb2:= Concatenation( SLAfcts.canbas( D, cg2 ) );
              mapi:= AlgebraHomomorphismByImagesNC( D, D, cb1, cb2 );
              map:= mapi*map;
           fi;
        
           pieces:= [ ];
           for i in [1..Length(dd)] do
               f0:= AlgebraHomomorphismByImagesNC( dd[i], dd[i], Basis(dd[i]),
                              List( Basis(dd[i]), x -> Image( map, x ) ) );
               t:= IsInnerAut_simple_real(f0);
               if t.isinn then
                  Add( pieces, f0 );
               else
                  c1:= List( cg1, u -> Filtered( u, x -> x in dd[i] ) );
                  inds:= List( c1[1], x -> Position( cg1[1], x ) );
                  tp:= CartanType( CartanMatrix(R){inds}{inds} );
                  type:= tp.types[1][1];
                  en:= tp.enumeration[1];
                  if type = "A" then
                     perm:= [ MappingPermListList( en, Reversed(en) ) ];
                  elif type ="D" then
                     if Length( en ) >= 5 then
                        perm:= [ MappingPermListList( en,
                                 Permuted(en,(Length(en)-1,Length(en))) ) ];
                     else
                        perm:= [MappingPermListList(en,Permuted(en,(3,4))),
                              MappingPermListList( en, Permuted(en,(1,3)) ),
                              MappingPermListList( en, Permuted(en,(1,4)) ),
                              MappingPermListList( en, Permuted(en,(1,3,4)) ),
                              MappingPermListList( en, Permuted(en,(1,4,3)) )];
                     fi;
                  else
                      perm:= [ MappingPermListList( en,
                                              Permuted(en,(1,6)*(3,5)) ) ];
                  fi;
                  c1:= CanonicalGenerators( R );
                  b1:= Concatenation( SLAfcts.canbas( dd[i], c1 ) );
                  for pi in perm do 
                      c2:= List( c1, x -> Permuted( x, pi ) );
                      b2:= Concatenation( SLAfcts.canbas( dd[i], c2 ) );
                      f1:= AlgebraHomomorphismByImagesNC( dd[i], dd[i],
                                                             b1, b2 )*f0;
                      t:= IsInnerAut_simple_real(f1);
                      if t.isinn then
                         Add( pieces, f1 );
                         break;
                      fi;
                  od;

               fi;
           od;

           mts:= List( pieces, h -> conjmat(h) );
           a:= Product( mts );               

           hp:= a*( g*ComplexConjugate(a)*g^-1 )*h;

           # Now we work with the identity component G^0 and the
           # conjugation is given by x ---> (a*g) \bar{x} (a*g)^-1.

           # determine the H2 of the centre of the simply connected group:

            sc:= ScCentre( D, a*g );

      # now sc is a record with the folowing fields:
      # (denote the elements of the centre of the simply connected group
      # by z_1,...,z_n).
      # sc.conj : a list consisting of 1,..,n 
      #           if sc.conj[i]=j then the conjugate of z_i is z_j.
      # sc.cc : list of n lists, if the i-th list is [t1,...,tr] then
      #         z_i = h_1(t1)...h_r(tr). 
      # sc.scelms : the elements as matrices in the matrix realization of the
      #             simply connected group
      # sc.maxcpt : a maximally compact Cartan subalgebra of the Lie algebra
      #             of the simply connected group (so this is a matrix algebra).
      # sc.U : the Lie algebra of the simply connected group.
      # sc.UtoL : function mapping elements of U to elements of D.

          # the next line is for later use...
            dU:= DirectSumDecomposition( sc.U );

          # now we determine the Z^2 and B^2 of the centre by brute force...
        
            z2:= [];
            for i in [1..Length(sc.conj)] do
                if sc.conj[i] = i then Add( z2, i ); fi;
            od;

            b2:= [ ];
            for i in [1..Length(sc.cc)] do
                u1:= sc.cc[i];
                u2:= sc.cc[ sc.conj[i] ];
                u:= List( [1..Length(u1)], j -> u1[j]*u2[j] );
                Add( b2, Position( sc.cc, u ) );
            od;

            classes:= [ ];
            for i in [1..Length(z2)] do
                u1:= sc.cc[ z2[i] ];
                cls:= [ ];
                for j in [1..Length(b2)] do
                    u2:= sc.cc[ b2[j] ];
                    u:= List( [1..Length(u1)], j -> u1[j]*u2[j] );
                    Add( cls, Position( sc.cc, u ) );
                od;
                Add( classes, Set(cls) );
            od;
            classes:= Set( classes );
            h2:= List( classes, x -> x[1] );

            # map them to the group we are dealing with:

            imh2:= List( sc.cc{ h2 }, c -> Product( [1..Length(c)], j ->
                            ha( j, c[j] ) ) );

            # imC is the image of the centre of the simply connected group,
            # so it is the centre of the semisimple part of our group,
            # that is, of G^0. The full centre is this times the central
            # torus. 

            imC:=  List( sc.cc, c -> Product( [1..Length(c)], j ->
                           ha( j, c[j] ) ) );

            # We want to know whether hp^-1 in Z^2C is equivalent to an
            # element of imh2. If hp^-1 is equivalent to z in imh2, then
            # there is an s in C such that h^-1 = s gam*s z, or hp^-1 z^-1
            # is equal to s gam*s. We check this with z running over imh2.
            # (By gam*s we denote the conjugation of s.)

            found:= false;
            for i in [1..Length(imh2)] do

                c:= hp^-1*imh2[i]^-1;
                if centraltorus then
                   u:= List( imC, x -> valpars( rC,
                                c*( x*a*g*ComplexConjugate(x)*(a*g)^-1 )^-1 ) );
                else
                   u:= List( imC, x ->
                                c = x*a*g*ComplexConjugate(x)*(a*g)^-1 );
                fi;

                pos:= PositionProperty( u, x -> x <> fail );

                if pos <> fail then
                   s1:= imC[pos];
                   found:= true;
                fi;

                # We have found s1 in the component group of the centre such
                # that c = s1 gam*s1 t, where t lies in the maximal central
                # torus. If such s1 does not exist then hp^-1 is not equivalent
                # to imh2[i]. 
           

                if found and centraltorus then

                   # need to establish whether c (s1 gam*s1)^-1 is lies in
                   # B^2 T where T is the maximal central torus, given by the
                   # record rC. 

                   # First we compute the conjugation of T, and an
                   # isomorphism to a torus in "normal form", that is a
                   # product of standard tori.
                   Q:= rC.Q;
                   A:= rC.C^-1*a*g*ComplexConjugate( rC.C );
                   U:= A*Product( rC.mats )*A^-1;
                   diag:= List( [1..Length(U)], i -> U[i][i] );
                   v:= List( Q, s -> Product( List( [1..Length(s)], i ->
                                                           diag[i]^s[i])));
                   m:= Dimension(T);
                   inds:= List( rC.indets, x->
                                        ExtRepPolynomialRatFun(x)[1][1] );
                   tau:= NullMat(m,m);
                   for l in [1..m] do
                       u:= ExtRepNumeratorRatFun( v[l] );
                       for k in [1,3..Length(u[1])-1] do
                           pos:= Position( inds, u[1][k] );
                           tau[l][pos]:= tau[l][pos]+u[1][k+1];
                       od;
                       u:= ExtRepDenominatorRatFun( v[l] );
                       for k in [1,3..Length(u[1])-1] do
                           pos:= Position( inds, u[1][k] );
                           tau[l][pos]:= tau[l][pos]-u[1][k+1];
                       od;
                   od;

                   # the element lies in B^2 T iff the parameters corresponding
                   # to the subtori isomorphic to E (consisting of u\in C^*
                   # with upgam u = \bar{u}) are positive. If such a parameter
                   # is not positive then we set found to false again.
                   # Otherwise we find the element M of the group such that
                   # the elt we deal with is equal to M * upgam M. 

                   is:= isomTS( tau );
                   v:= valpars( rC,
                             c*( s1*(a*g)*ComplexConjugate(s1)*(a*g)^-1 )^-1 );
                   v1:= is.TtoS(v);
                   v:= [ ];
                   for k in [1..is.flen] do
                       v[k]:= Sqroot(v1[k]);
                   od;
                   for k in [is.flen+1..is.flen+is.elen] do
                       if ispositive(v1[k]) = false then
                          found:= false;
                          break;
                       elif ispositive(v1[k]) = true then
                          v[k]:= Sqroot(v1[k]);
                       else
                          Print("ERROR!!, not clear if ",v1[k]," is positive."); 
                       fi;
                   od;
                   if found then
                      # finish v
                      for k in [is.flen+is.elen+1..is.flen+is.elen+is.glen] do
                          v[k]:= v1[k];
                          v[k+is.glen]:= One(SqrtField);
                      od;

                      v:= is.StoT(v);
                      M:= rC.C*Product( rC.mats )*rC.C^-1;
                      for k in [1..Length(M)] do
                          for l in [1..Length(M)] do
                              M[k][l]:= Value( M[k][l], rC.indets, v );
                          od;
                      od;

                      s1:= M*s1;
                   fi;

                fi;
                if found then

                   rT:= toruspar_gen( SqrtField, sc.maxcpt );
                   v:= valpars( rT, sc.scelms[h2[i]] );
                   v:= List( v, Sqroot );
                   M:= rT.C*Product( rT.mats )*rT.C^-1;
                   for k in [1..Length(M)] do
                       for l in [1..Length(M)] do
                           M[k][l]:= Value( M[k][l], rT.indets, v );
                       od;
                   od;

                   if M = M^0 then
                      b:= s1;
                   else

                      # we have to map M to G^0...
                      M0:= Basis(D)[1]![1]^0;
                      for k in [1..Length(dU)] do
                          f0:= AlgebraHomomorphismByImagesNC( dU[k], dU[k],
                            Basis( dU[k] ), List( Basis(dU[k]), x -> M*x*M^-1));
                          wd:= inneraut_asword( f0 );
                          Rd:= RootsystemOfCartanSubalgebra( dU[k],
                                MaximallyNonCompactCartanSubalgebra(dU[k]) );
                          ch:= ChevalleyBasis(Rd);
                          rv:= Concatenation( ch[1], ch[2] );
                          M0:= M0*Product( wd, z->xa( sc.UtoL(rv[z[1]])![1],
                                  z[2] ) );
                      od;

                      b:= s1*M0;
                   fi;

                   d0:=b*a;
                   break;
                fi;

                   
            od;

            if not found then
               return fail;
            else
               return d0;
            fi;

        fi;

end;

neutralize2coc_non_reductive:= function( L, D, T, U, R, rC, g )

        # L: the Lie algebra of the group
        # D: the derived algebra (so the semisimple part),
        # T: the centre of the reductive part
        # U: the nilpotent radical
        # R: root sys of D wrt max cpt Cartan
        # rC: the output of toruspar_gen with input T
        # g: the element so that we need to trivialize [h,f] where
        # h = g (gam*g), f = inn(g)\circ \sigma. 

        local sig, S1, S2, uu, s, g1, d1, d2, g2, h, g1p;

        if Dimension(D)=0 and Dimension(T)=0 then
           h:= g*ComplexConjugate(g);
           return  _exp( (-1/2)*_log(h) );
        fi;

        sig:=function( v ) return LieObject(ComplexConjugate(v)); end;

        S1:= Subalgebra( L, Concatenation(Basis(D),Basis(T)), "basis" );
        S2:= Subalgebra( L, List( Basis(S1), x -> g*sig(x)*g^-1 ), "basis" );

        uu:= conjugatereductive( L, S2, S1, U );

        if Length(uu) = 0 then
           s:= g^0;
        else
           s:= Product( List( uu, x -> _exp(x) ) );
        fi;

        g1:= (s*g)*ComplexConjugate( s*g );
        g1p:= projection( L, D, T, U, g1 );

        d1:= neutralize2cocreductive( D, T, R, rC, g1p );

        if d1=fail then return fail; fi;
        
        g2:= d1*s*g;
        h:= g2*ComplexConjugate(g2);

        # now h lies in the unipotent radical...

        d2:= _exp( (-1/2)*_log(h) );

        return d2*d1*s;
end;

normalcc:= function( r, L, S, T, N, c)

        local Q, cr, sr, gi, u, si;

        if Dimension(N) = 0 then
           if Dimension(S) = 0 then
              return normalize_toruscoc( r, c );
           else
              return normalizecoc( r, c );
           fi;
        else

           if Dimension(S)=0 and Dimension(T)=0 then
              return _exp( (1/2)*_log(c) );
           fi;

           Q:= r.QQ; 
           cr:= projection( L, S, T, N, c );
           if Dimension(S) = 0 then
              sr:= normalize_toruscoc( r, cr );
           else
              sr:= normalizecoc( r, cr );
           fi;
           gi:= sr^-1*cr*Q*ComplexConjugate(sr)*Q^-1;
           u:= sr^-1*c*Q*ComplexConjugate(sr)*Q^-1*gi^-1;

           si:= _exp( (1/2)*_log(u) );
           return sr*si;
        fi;
end;
      

H1disconnected:= function( H1, L, tab, gam, pinv )

        # everything over SqrtField

        # here:

        # H1: an H1DataObj, storing lots of information needed for
        #     normalization of cocycles
        # mats - matrices forming a basis of the Lie algebra
        # tab - a multiplication table of the component group,
        #       so mat[i][j]=k means g_i*g_j = g_k 1 <= i,j,k <= n
        # gam - is a function giving the conjugation on the component group,
        #       so the conjugate of i is gam(i)
        # pinv - is a list of elements of the group giving the inverse
        #        images of the elements of the component group, so
        #        for 1 <= i <=n we have that pinv[i] is an inverse image
        #        of i. 
        
        local n0, one, inv, prod, D, T, rC, sig, TD, R, cg1, cb1, H1C, ld,
              HH1, q, g, h, g0, found, a, h1, Ctau, act, k, l, b, c0, orbs, U,
              mats0, elts, oo, inds, i, sp, D0, T0;

        n0:= Length( tab );
        one:= Position( tab, [1..n0] );

        inv:= function( i )
               return Position( tab[i], one );
        end;

        prod:= function( i, j )
              return tab[i][j];
        end;

        H1!.liealg:= L;


        ld:= levidecomposition( L, function(x) return
                  LieObject(ComplexConjugate(x![1]) ); end);
        R:= ld[1]; U:= ld[2];

        mats0:= List( Basis(R), x -> x![1] );

        D:= LieDerivedSubalgebra( R );
        T:= LieCentre(R);

        H1!.liealg:= L;
        H1!.levi:= D;
        H1!.centraltor:= T;
        H1!.unipot:= U;

        if Dimension( T ) > 0 then
           rC:= toruspar_gen( SqrtField, List( Basis(T), x -> x![1] ) );
        else
           rC:= fail; 
        fi;

        if Dimension(D) > 0 then

           sig:= function( v ) return LieObject(ComplexConjugate(v)); end;
# later use more general conjugation...
           SetRealStructure( D, sig );
           TD:= MaximallyNonCompactCartanSubalgebra( D );
           R:= RootsystemOfCartanSubalgebra( D, TD );

           cg1:= CanonicalGenerators( R );
           cb1:= Concatenation( SLAfcts.canbas( D, cg1 ) );
           R!.canbas1:= cb1;

        else
           R:= fail;
        fi;

        H1C:= H1finite( tab, gam );
        H1!.h1_compgrp:= H1C;
        H1!.preims_cocs:= [ ];
        H1!.connectedH1s:= [ ];
        H1!.Ctaus:= [ ];
        H1!.orbs:= [ ];
        H1!.actingelts:= [ ];
        H1!.levis:= [ ];

        HH1:= [ ];
        for q in H1C do

            g:= pinv[q];
            h:= g*ComplexConjugate(g);   # later use more general conjugation...

            if h = h^0 then # g already is a cocycle...
if printstuff then 
   Print("H1disconnected: g is a cocycle!!\n");
fi;
               g0:= g;
               Add( H1!.preims_cocs, g );
               found:= true;

            else # now we need to modify g so that it becomes a cocycle,
                 # or decide that this is not possible...
if printstuff then 
   Print("H1disconnected: g is not a cocycle!!\n");
fi;

               if Dimension(U) = 0 then
                  a:= neutralize2cocreductive( D, T, R, rC, g );
               else
                  a:= neutralize2coc_non_reductive( L, D, T, U, R, rC, g );
               fi;

               if a = fail then
                  found:= false;
                  Add( H1!.preims_cocs, fail );
               else
                  found:= true;
                  g0:= a*g;
                  Add( H1!.preims_cocs, g0 );
if printstuff then 
   Print("H1disconnected: u is a cocycle... ",g0*ComplexConjugate(g0)=g0^0,"(should be true)\n");
fi;

               fi;
            fi;

            if found then


               if Dimension(D)=0 and Dimension(T)=0 then

                  h1:= rec( cocyc:= [
                        IdentityMat( Length(Basis(U)[1]![1]),
                                     LeftActingDomain(U) ) ] );
                  D0:= D; T0:= T;
               else

                  # check whether the reductive part is stable under the new
                  # conjugation
                  sp:= VectorSpace( LeftActingDomain(L), mats0 );
                  if not ForAll( mats0,
                       x -> g0*ComplexConjugate(x)*g0^-1 in sp ) then

                       ld:= levidecomposition( L,
                         function(x) return
                           LieObject( g0*ComplexConjugate(x![1])*g0^-1 ); end);
                       R:= ld[1]; 
                       D0:= LieDerivedSubalgebra( R );
                       T0:= LieCentre(R);
                       h1:=  H1connected( List( Basis(R), x -> x![1] ), g0 );
                  else
                       h1:= H1connected( mats0, g0 );
                       D0:= D; T0:= T;
                  fi;

               fi;

# THE input here should be a Lie algebra as this is called many times...
# Yes, but the question is whether the real structure of the Lie algebra is
# set... if yes, then we need to make a copy in H1connected, or so...

               Add( H1!.connectedH1s, h1 );
               Add( H1!.levis, [D0,T0] );

# THIS has to be changed if the conjugation does not preserve D+T, where the
# conjugation is x --> g0 (gam*x) g0^-1. In that case we have to compute new
# D, T and put them in the list.

               # now we determine C^\tau...
               # q is the cocycle by which we twist in C...

               Ctau:= Filtered( [1..n0], k ->
                           prod( q, prod( gam(k), inv(q) ) ) =k );

               Add( H1!.Ctaus, Ctau );

               # now we compute the action of Ctau on h1...

               act:= List( [1..Length(h1.cocyc)], x -> [ ] );
               elts:= List( [1..Length(h1.cocyc)], x -> [ ] );

               # act[k] contains the images of h1.cocyc[k] under the elts
               # of Ctau...

# THE next bit can be more efficient: if k already lies in an orbit then
# we don't need to consider all images of k!!

               for k in [1..Length(act)] do
                   for l in [1..Length(Ctau)] do
                       b:= pinv[ Ctau[l] ];
                       c0:= b^-1*h1.cocyc[k]*(g0*ComplexConjugate(b)*g0^-1);
if printstuff then 
   Print("H1disconnected: c0 is cocycle... ",c0*(g0*ComplexConjugate(c0)*g0^-1)=c0^0," (should be true)\n");
fi;

                       g:= normalcc( h1, L, D0, T0, U, c0 );

                       c0:= g^-1*c0*(g0*ComplexConjugate(g)*g0^-1);
                       act[k][l]:= Position( h1.cocyc, c0 );
                       elts[k][l]:= g;
                   od;
               od;

               oo:= List( act, Set );
               orbs:= [ ];
               inds:= [ ]; # the indices of the final orbits
               for i in [1..Length(oo)] do
                   if not oo[i] in orbs then
                      Add( orbs, oo[i] );
                      Add( inds, i );
                   fi;
               od;

               Add( H1!.orbs, act{inds} );
               Add( H1!.actingelts, elts{inds} );

               Append( HH1, List( inds, x -> h1.cocyc[x]*g0) );

            else
               Add( H1!.connectedH1s, fail );
               Add( H1!.levis, fail );
               Add( H1!.orbs, fail );
               Add( H1!.actingelts, fail );
            fi;

        od;

        return HH1;
end;

iseltof:= function( L, g )


        # here L is a matrix Lie algebra,in gl(n,C)
        # it is the Lie algebra of the connected algebraic group
        # G\subset GL(n,C),
        # the function returns true if g in G, false otherwise

        local F, jd, s, u, u1, u2, lgu, i, S, cgs, H, basT, T, r;

        F:= LeftActingDomain(L);

        if not IsBound( L!.istorus ) then

           if IsLieAbelian(L) then
             if ForAll(Basis(L),x->x![1]=JordanDecomposition(x![1])[1]) then
                L!.istorus:= true;
             else
                L!.istorus:= false;
             fi;
           else
             L!.istorus:= false;
           fi;
        fi;

        if L!.istorus then
           if not IsBound( L!.toruspar ) then
              L!.toruspar:= toruspar_gen( F, List( Basis(L), x -> x![1] ) );
           fi;
           r:= L!.toruspar;
           return valpars( r, g ) <> fail;           
        fi;
        
        # We compute the multiplicative Jordan decomposition:

        jd:= JordanDecomposition( g );
        s:= jd[1];
        u:= s^-1*g;

        u1:= u-u^0;
        lgu:= Zero(F)*u;
        u2:= u1;
        i:= 1;
        while not IsZero(u2) do
            lgu:= lgu+(-1)^(i-1)*u2/i;
            u2:= u2*u1; i:= i+1;
        od;

        if not LieObject(lgu) in L then
           return false;
        fi;

        if IsOne(s) then return true; fi;

        S:= List( Basis(L), x -> Coefficients( Basis(L), s*x*s^-1-x ) );

        if fail in S then return false; fi;

        cgs:= Subalgebra( L, List( NullspaceMat( S ), x -> x*Basis(L) ) );

        H:= CartanSubalgebra( cgs );

        basT:= List( Basis(H), x -> LieObject(JordanDecomposition(x![1])[1]));
        T:= Subalgebra( L, basT );

        if Dimension(T) = 0 then return false; fi;

        r:= toruspar_gen( F, List( Basis(T), x -> x![1] ) );
        return valpars( r, s ) <> fail;

end;


FirstGaloisCohomology:= function( arg )

        local r, mats, tab, z, gam, pinv, one, H1, L, ld, str, R, tp, i,
              incomp, j;

        one:= One( SqrtField );

        H1:= NewH1Obj();
        if Length(arg)=1 then

           L:= LieAlgebra( SqrtField, arg[1]*one );
           ld:= levidecomposition( L, 
                  function(x) return LieObject(ComplexConjugate(x![1])); end);

           r:= H1connected( List( Basis(ld[1]), x -> x![1] ) );
# need to change that, so as to input the Lie algebra directly!
           H1!.redcondata:= r;
           H1!.isconnected:= true;
           H1!.liealg:= L;
           H1!.unipot:= ld[2];
           H1!.levi:= LieDerivedSubalgebra( ld[1] );
           H1!.centraltor:= LieCentre( ld[1] );
           SetCocycles( H1, r.cocyc );
        else
           mats:= arg[1]*one;
           L:= LieAlgebra( SqrtField, mats );
           pinv:= arg[2]*one;
           incomp:= function( g ) 

               local pos, i;

               pos:= Position( pinv, g );
               if pos <> fail then return pos; fi;

               for i in [1..Length(pinv)] do 
                   if iseltof( L, pinv[i]^-1*g ) then
                      return i; 
                   fi;
               od;
               return fail; 
           end;
                   
           tab:= List( pinv, a -> [] );
           for i in [1..Length(pinv)] do 
               for j in [1..Length(pinv)] do 
                   tab[i][j]:= incomp( pinv[i]*pinv[j] );
               od;
           od;
           z:= List([1..Length(pinv)], i -> incomp(ComplexConjugate(pinv[i])));
           gam:= function( i ) return z[i]; end;
           H1!.isconnected:= false;
           H1!.inputdata:= [ L, tab, gam, pinv ];
           SetCocycles( H1, H1disconnected( H1, L, tab, gam, pinv ) );
        fi;

        str:="";
        if Dimension(H1!.levi)>0 then
           R:= RootSystem( H1!.levi );
           tp:= CartanType( CartanMatrix(R) );
           if Length(tp.types) > 1 then Append( str, "( " ); fi;
           for i in [1..Length(tp.types)] do 
               if i > 1 then Append( str, " + " ); fi; 
               Append( str, tp.types[i][1] );
               Append( str, "_" );
               Append( str, String( tp.types[i][2] ) );
           od; 
           if Length(tp.types) > 1 then Append( str, " )" ); fi;
           if Dimension(H1!.centraltor) > 0 then
              Append( str, ".T_" );
              Append( str, String(Dimension(H1!.centraltor)) );
           fi;
           if Dimension(H1!.unipot) > 0 then
              Append( str, "|x U_" );
              Append( str, String( Dimension(H1!.unipot) ) );
           fi;
        elif Dimension(H1!.centraltor) > 0 then
              Append( str, "T_" );
              Append( str, String(Dimension(H1!.centraltor)) );
              if Dimension(H1!.unipot) > 0 then
                 Append( str, "|x U_" );
                 Append( str, String( Dimension(H1!.unipot) ) );
              fi;
        else
              Append( str, "U_" );
              Append( str, String( Dimension(H1!.unipot) ) );
        fi;

        if H1!.isconnected then 
           Append( str, ", connected; ");
        else
           Append( str, ", " );
           Append( str, String( Length(pinv) ) );
           Append( str, " components; " );
        fi;

        Append( str, String( Length( Cocycles(H1) ) ) );
        Append( str, " cocycles");

        H1!.printstr:= str;

        return H1;

end;

NormalizeCocycle:= function( H1, c )

        # here H1 is an H1DataObj, and c a cocycle in the corresponding
        # group G; we return a g in G such that g^-1*c*\bar{g} = c_i,
        # where c_i is the i-th element in the list Cocycles( H1 ).

        local r, N, S, T, L, tab, gam, n, one, inv, prod, pinv, i, k0,
              h1f, j, i0, j0, s, gi, h, b, k, pos, l, g1;

        if H1!.isconnected then 
           r:= H1!.redcondata;
           N:= H1!.unipot;
           S:= H1!.levi;
           T:= H1!.centraltor;
           L:= H1!.liealg;
           return normalcc( r, L, S, T, N, c );

        else

           tab:= H1!.inputdata[2];
           gam:= H1!.inputdata[3];

           n:= Length( tab );
           one:= Position( tab, [1..n] );

           inv:= function( i )
                 return Position( tab[i], one );
           end;

           prod:= function( i, j )
               return tab[i][j];
           end;

           L:= H1!.liealg;
           pinv:= H1!.inputdata[4];
           for i in [1..Length(pinv)] do
               if iseltof( L, c*pinv[i] ) then
                  k0:= i;
                  break;
               fi;
           od;
            
           h1f:= H1!.h1_compgrp;

           for j in [1..n] do 
               i:= Position( h1f, prod( inv(j), prod( k0, gam(j) ) ) );
               if i <> fail then
                  i0:= i; j0:= j;
                  break;
               fi;
           od;

           s:= pinv[j0];
           gi:= H1!.preims_cocs[i0];

           h:= s^-1*c*ComplexConjugate(s)*gi^-1;

           r:= H1!.connectedH1s[i0];

           b:= normalcc( r, L, H1!.levis[i0][1], H1!.levis[i0][2], H1!.unipot, h);

           k:= Position( r.cocyc, b^-1*h*(gi*ComplexConjugate(b)*gi^-1) );

           pos:= PositionProperty( H1!.orbs[i0], u -> k in u );

           l:= Position( H1!.orbs[i0][pos], k );
           g1:= H1!.actingelts[i0][pos][l];

           return s*b*g1^-1;

        fi;
end;


# exa:
#K:= SimpleLieAlgebra("C",3,Rationals);
#no:= NilpotentOrbits(K);
#C:= LieCentralizer( K, Subalgebra( K, [ SL2Triple(no[3])[3] ] ) );
#mats:= List( Basis(C), x -> AdjointMatrix(Basis(K),x) );;
#h1:= FirstGaloisCohomology( mats );        
