#########################################################################
#
# this file contains the methods for defining the field SqrtField
# and for working with its elements, see sqrt.gd
#
#

###############################################
# Do deal with Flat for lists of matrices
#
#corelg.myflat := function(L)
#   return Concatenation(L);
#end;

# start sage...
sage_proc:= InputOutputLocalProcess( DirectoryCurrent(),
"/home/degraaf/software/SageMath/sage", [] );


WriteLine( sage_proc, "load(\'/home/degraaf/gapprog/galcohom/sagefct.sage\')" );


readsage:= function()

     local s, str;

     WriteLine( sage_proc, "outputTerm()" );
     s:="";
     repeat    
         str := ReadAll( sage_proc );
	 s:= Concatenation( s, str );
     until PositionSublist( str, "@-----@" ) <> fail;
     return s;
end;

#str:= readsage();

###############################################
# DUE TO A BUG IN BasisNC
InstallMethod( NiceBasisNC, 
               [IsBasisByNiceBasis and HasNiceBasis], NiceBasis );

SqrtFieldFam!.newsq:=[ ]; 
SqrtFieldFam!.newstr:=[ ];
SqrtFieldFam!.newp:=[ ];
SqrtFieldFam!.newconj:=[ ]; # contains the complex conjugates in extrep


# if newsq[i] = list and newp[i] = p then list is the extrep of a and
# the element with index -i is pRoot( p, a )
# in that case newstr[i] is the string version of pRoot( p, a )

# in an extrep the p-th roots are reresented by [ -i, k ] meaning
# pRoot( p, a )^k

#SqrtFieldFam!.newbas:=[ ];
#SqrtFieldFam!.newEn:=[ ];

SQRTS:= [  ];
SQRTINDS:= [ ];
# if SQRTINDS[i] = k and k > 0 then Sqroot(k) = a_i
# if k < 0 then the element with index k is a_i
# (this is then a more complicated root). Here a_i = SQRTS[i].

WriteLine( sage_proc, "K.<a_0>=CyclotomicField(4)" );
WriteLine( sage_proc, "emb=K.complex_embeddings(100)[0]" );
str:= readsage();
ROOT:= 0;

# here newsq has the new elts from which we take the square root,
# newstr has the corresponding strings for printing,
# newbas has the basis as extension of the old SqrtField,
# newEn has the sequances of 0,1 that correspond to this basis.

# small constant for deciding whether certain things are "close enough" to 0.
eps_0:= 0.001;


#########################################################################

BindGlobal( "SqrtField_objectify", function( elm )
local u,isrt;
   isrt:= Length(elm) = 1 and Length(elm[1][2])=0;
   u:= Objectify( SqrtFieldType, [ Immutable(elm) ] );
   u![2]:= isrt;
   return u;
end ); 

#########################################################################

InstallMethod( \in,
   "for an object and a SqrtField",
   [ IsObject, IsSqrtField ], 1000000,
function( p, qf )
   return IsSqrtFieldElement(p);
end );

#########################################################################

sfelt_to_string:= function( elt ) # here elt is the ext rep...

   local l, printmon, st, i, start;

   l:= Length(elt);
   printmon := function(mon)
      local st, pos, tail, k, start;
#   Print("mon...  ",mon,"\n");   
      pos:= PositionProperty( mon[2], x -> IsList(x) );
      if pos = fail then pos:= Length(mon[2])+1; fi;
      tail:= "";
      for k in [pos..Length(mon[2])] do
          Append( tail, SqrtFieldFam!.newstr[-mon[2][k][1]] );
	  if mon[2][k][2] > 1 then
	     Append( tail, Concatenation( "^", String(mon[2][k][2]) ) );
	  fi;
          if k < Length(mon[2]) then Append( tail, "*" ); fi;
      od;
      if pos > 1 then
         start:= Concatenation( "Sqroot( ",String(Product(mon[2]{[1..pos-1]})),
	      " )" );
      else
         start:= "";
      fi;
      if mon[1] = 1 then
         st:= start;
	 if Length(tail) > 0 then
	    if Length(st) > 0 then 
  	       st:= Concatenation( st, "*", tail );
	    else
	       st:= tail;
	    fi;
	 fi;
      elif mon[1]<0 or not (IsRat(mon[1]) or IsRat(E(4)*mon[1])) 
                    or (IsRat(E(4)*mon[1]) and E(4)*mon[1]>0) then
         st:= Concatenation( "(",String(mon[1]),")*", start);
	 if Length(tail) > 0 then
	    st:= Concatenation( st, "*", tail );
	 fi;
      else     
         st:= Concatenation( String(mon[1]),"*",start);
	 
	 if Length(tail) > 0 then
	    if Length(start) > 0 then 
  	       st:= Concatenation( st, "*", tail );
	    else
	       st:= Concatenation( st, tail );
	    fi;
	 fi;
      fi;
#Print(st,"\n");      
      return st;
   end;
   if l=1 then
      if elt[1][2]=[] then
         return String(elt[1][1]);
      else
         return printmon(elt[1]);
      fi;
   else
      st:= "";
      if elt[1][2]=[] then
         Append( st, String(elt[1][1]) );
	 Append( st, " + " );
      else 
         Append( st, printmon(elt[1]) ); 
         Append(st, " + "); 
      fi;
   fi;
   for i in [2..l-1] do
       Append( st, printmon(elt[i]) );
       Append( st, " + ");
   od;
   if l>1 then Append( st, printmon(elt[l]) ); fi;  
   return st;
end;

sfelt_to_sage_string:= function( elt ) # here elt is the ext rep...

   local printmon, j, str, s, exp;

   printmon := function(mon)
      local st, pos, tail, k, i, start, cf;
#   Print("mon...  ",mon,"\n");

      if mon[1]=1 and Length(mon[2])=0 then
         return "1";
      fi;
      
      st:= "";
      for k in [1..Length(mon[2])] do
          if IsList( mon[2][k] ) then
	     s:= mon[2][k][1];
	     exp:= mon[2][k][2];
	  else
	     s:= mon[2][k];
	     exp:= 1;
	  fi;
          i:= Position( SQRTINDS, s );
	  Append( st, Concatenation("a_",String(i) ) );
	  if exp > 1 then
	     Append( st, Concatenation( "^", String(exp) ) );
	  fi;
          if k < Length(mon[2]) then Append( st, "*" ); fi;
      od;

      if mon[1] = 1 then
         start:= "";
      else
         start:= "(";
         cf:= Coefficients( Basis(CF(4)), mon[1] );
         if cf[1] <> 0 and cf[2]<> 0 then
	    start:= Concatenation( start, String(cf[1]),"+",String(cf[2]),
	                  "*a_0");
	 elif cf[1] <> 0 then
	    start:= Concatenation( start, String(cf[1]) );
	 elif cf[2] <> 0 then
	    start:= Concatenation( start, String( cf[2] ),"*a_0");
	 fi;
	 start:= Concatenation(start, ")" );
      fi;

      if Length( start ) > 0 and Length( st ) > 0 then
       	 st:= Concatenation( start, "*", st  );
      elif Length(start) > 0 then
         st:= start;
      fi;

#Print(st,"\n");      
      return st;
   end;

   str:= "";
   for j in [1..Length(elt)] do 
       str:= Concatenation( str, printmon( elt[j] ) );
       if j < Length( elt ) then
          str:= Concatenation(str,"+");
       fi;
   od;

   return str;
   
end;

InstallGlobalFunction(Sqroot, function(arg)
local d, n, fc, cf, ps, i, m, sgn, inv, found, k, u, eu, a, b, s, t, p, primes,
      e, mon, str, q, exp, strq, pos, j;

   if Length(arg)=1 then
      q:= arg[1]; exp:= 2;
   else
      exp:= arg[1]; q:= arg[2];
   fi;

   if not (IsSqrtFieldElement(q) or IsRat(q)) then
      Error("<q> must be a rational, or rational SqrtField elt"); 
   fi;
   
   if IsSqrtFieldElement(q) and Length(q![1]) = 1 and IsRat(q![1][1][1])
      and q![1][1][2] = []  then
      return Sqroot(exp,q![1][1][1]);
   fi;


   if IsSqrtFieldElement(q) or exp > 2 then

      if IsSqrtFieldElement(q) then
         pos:= Position( SqrtFieldFam!.newsq, q![1] );
	 if pos <> fail and SqrtFieldFam!.newp[pos] = exp then
	    return SqrtField_objectify( [ [ 1, [ [-pos, 1] ] ] ] );
	 fi;
         strq:= sfelt_to_sage_string( q![1] );
      else
         pos:= Position( SqrtFieldFam!.newsq, q );
	 if pos <> fail and SqrtFieldFam!.newp[pos] = exp then
	    return SqrtField_objectify( [ [ 1, [ [-pos, 1] ] ] ] );
	 fi;
         strq:= String(q);
      fi;

      WriteLine( sage_proc, "R.<y> = PolynomialRing( K )" );
      str:= readsage();
      WriteLine( sage_proc, Concatenation( "f = y^",String(exp),"-(",
	         strq,")" ) );

      str:= readsage();

      WriteLine( sage_proc, Concatenation( "printGAP( f, ",
	                 String(Length(SQRTS)), ")" ) );
      str:= readsage();

      Read("tmp.g");
      found:= true;
#Print(ROOT,"\n");
      if ROOT = 0 then
	 found:= false;
      else
	 return ROOT;
      fi;
      if not found then
         if IsRat(q) then
	    Add( SqrtFieldFam!.newsq, q );
	    strq:= String(q);
	 else   
     	    Add( SqrtFieldFam!.newsq, q![1] );
	    strq:= sfelt_to_string( q![1] );
	 fi;
	 Add( SqrtFieldFam!.newstr, Concatenation( "pRoot( ",
	                  String(exp), ",", strq, " )" ) );
         Add( SqrtFieldFam!.newp, exp );			  
	 u:= SqrtField_objectify([[1,[[-Length(SqrtFieldFam!.newsq),1]]]]);
	 Add( SQRTS, u );
	 Add( SQRTINDS, -Length(SqrtFieldFam!.newsq) );

         # extend the Sage field...
	 i:= Length( SQRTS );
	 WriteLine( sage_proc, Concatenation( "K.<a_",String(i),
	             ">=NumberField( f )" ) );
         str:= readsage();
	 str:= "[a_0";
	 for j in [1..i-1] do
	     Append( str, Concatenation( ",a_",String(j) ) );
	 od;
	 Append( str, "]" );
	 WriteLine( sage_proc, Concatenation( "emb = newEmb( K, ", str,
	        ",emb )" ) );
	 str:= readsage();
	 return u;
      fi;
   fi;
   
   if q < 0 then sgn := E(4); q := -q; else sgn := 1; fi;
   if q in [0,1] then 
      return SqrtField_objectify([[sgn*q,[]]]); 
   fi;
   d  := DenominatorRat(q);
   n  := NumeratorRat(q);
   fc := FactorsInt(n);
   Append(fc,FactorsInt(d));
   fc := Collected(fc);
   cf := 1/d;
   ps := [];
   for i in [1..Length(fc)] do
      m  := fc[i][2] mod 2;
      cf := cf *(fc[i][1]^((fc[i][2]-m)/2));
      if m=1 then Add(ps,fc[i][1]); fi;
   od;
   if ps = [] then return 
      SqrtField_objectify([[sgn*cf,[]]]); 
   fi; 
   if ps[1] = 1 then ps := ps{[2..Length(ps)]}; fi;
   Sort(ps);
   MakeImmutable(ps);   
   for i in [1..Length(ps)] do
       u:= SqrtField_objectify([[1,[ps[i]]]]);
       if not u in SQRTS then
          Add( SQRTS, u );
	  Add( SQRTINDS, ps[i] );
	  WriteLine( sage_proc, "R.<y>=PolynomialRing( K )" );
	  str:= readsage();
	  WriteLine( sage_proc, Concatenation( "f=y^2-",String(ps[i]) ) );
	  str:= readsage();
	  i:= Length(SQRTS);
          WriteLine( sage_proc, Concatenation( "K.<a_",String(i),
	             ">=NumberField( f )" ) ); 
          str:= readsage();
	  str:= "[a_0";
	  for j in [1..i-1] do
	     Append( str, Concatenation( ",a_",String(j) ) );
	  od;
	  Append( str, "]" );
	  WriteLine( sage_proc, Concatenation( "emb = newEmb( K, ", str,
	        ",emb )" ) );
	  str:= readsage();
       fi;
   od;
   return SqrtField_objectify([[sgn*cf,ps]] );
end);

nRoot:= function( exp, q )

    local fc, a, f;

    if exp=1 then return q; fi;

    fc:= Factors( exp );
    a:= q;
    for f in fc do
	if not IsSqrtFieldElement(a) then a:= a*One(SqrtField); fi;
        a:= Sqroot( f, a );
    od;

    return a;
end;    

#########################################################################

SetOne( SqrtField, Sqroot(1) );
SetZero( SqrtField, Sqroot(0) );
SetZero( SqrtFieldFam, Zero(SqrtField) );
SetOne( SqrtFieldFam, One(SqrtField) );
SetIsUFDFamily(SqrtFieldFam,true);


########################################################################

rcp_0:= [];
approx:= function( q )

   local strq, str;

   if not IsSqrtFieldElement(q) then
      Error("not a SqrtElt!!");
   fi;
   strq:= sfelt_to_sage_string( q![1] );
   WriteLine( sage_proc, Concatenation( "value(", strq, ")" ) );
   str:= readsage();
   Read("tmp.g");   
   return rcp_0;
end;

ispositive:= function( q )

   # here q is a sqrt field element that we assume to be real,
   # we return true if it is >0 (i.e., > eps_0), false if it is < 0
   # (i.e., < -eps_0), and fail otherwise.

   local u;

   u:= approx(q)[1];
   if u > eps_0 then
      return true;
   elif u < eps_0 then
      return false;
   else
      return fail;
   fi;


end;

#########################################################################

InstallMethod( ViewObj, 
        "for SqrtField elements",
        [ IsSqrtFieldElement ],

function( sf )
    Print( sfelt_to_string( sf![1] ) );
end ); 

##########################################################################

InstallMethod( PrintObj, 
        "for SqrtField elements",
        [ IsSqrtFieldElement ],
function( sf )
    Print( sfelt_to_string( sf![1] ) );
end );

#########################################################################


InstallGlobalFunction( SqrtFieldIsGaussRat, function(el)
   return IsSqrtFieldElement(el) and el![2];
end); 

#########################################################################


InstallGlobalFunction( SqrtFieldMakeRational, function(el)
   if IsSqrtFieldElement(el) then
      if SqrtFieldIsGaussRat(el) then
         return el![1][1][1]; 
      else 
         return false; 
      fi;
   elif IsSqrtFieldElementCollection(el) then
      if ForAll(el,SqrtFieldIsGaussRat) then 
         return List(el,x->x![1][1][1]);
      else
         return false;
      fi;
   elif IsSqrtFieldElementCollColl(el) then
      if ForAll(Flat(el),SqrtFieldIsGaussRat) then
         return List(el,x->List(x,y->y![1][1][1]));
      else
         return false;
      fi; 
   fi;
   return false;
end); 

#########################################################################

InstallMethod( Determinant,
   "for SqrtField matrices",
   [ IsSqrtFieldElementCollColl], 1000,
function(x)
local a,b;
   a := SqrtFieldMakeRational(x);
   if not a = false then 
      return DeterminantMat(a)*One(SqrtField);
   fi;
   TryNextMethod();
end);

#########################################################################

InstallMethod( \*,
   "for SqrtField matrices",
   [ IsSqrtFieldElementCollColl, IsSqrtFieldElementCollColl],
function(x,y)
local a,b;
   a := SqrtFieldMakeRational(x);
   if not a = false then 
      b := SqrtFieldMakeRational(y);      
      if not b = false then
         return One(SqrtField)*(a*b);
      fi;
   fi;
  #Print("DONT USE MAKERAT\n");
  TryNextMethod();
end);


#########################################################################

InstallMethod( \+,
   "for SqrtField matrices",
   [ IsSqrtFieldElementCollColl, IsSqrtFieldElementCollColl],
function(x,y)
local a,b;
   a := SqrtFieldMakeRational(x);
   if not a = false then 
      b := SqrtFieldMakeRational(y);      
      if not b = false then
         return One(SqrtField)*(a+b);
      fi;
   fi;
  #Print("DONT USE MAKERAT\n");
   TryNextMethod();
end);


#########################################################################
#
#InstallMethod( \*,
#   "for rational and SqrtField elements",
#   [IsCyclotomic, IsSqrtFieldElement], 
#function(x,y)
#local b,i;
#      if not IsGaussRat(x) then
#         Error("cyclotomic has to be Gaussian Rational");
#      fi;
#      if x = 0 then return Zero(SqrtField); fi;
#      if x = 1 then return y; fi;
#      b := List(y![1],ShallowCopy);
#      for i in b do i[1] := x*i[1]; od;
#      return SqrtField_objectify( b );          
#end);

PRU:= function( k ) # primitive root of unity in SqrtField...

        local a,z;

        a:= Sqroot( -10 + 2*Sqroot( 5 ) );

        z:= (1/16-1/16*E(4))*Sqroot( 2 ) + (-1/16-1/16*E(4))*Sqroot( 6 ) +
            (1/16+1/16*E(4))*Sqroot( 30 )+
            (1/32+1/32*E(4))*Sqroot( 30 )*a +
            (1/32+1/32*E(4))*Sqroot( 6 )*a +
            (-1/16+1/16*E(4))*Sqroot( 10 ) +
            (-1/32+1/32*E(4))*Sqroot( 10 )*a +
            (-1/32+1/32*E(4))*Sqroot( 2 )*a;


        if k=1 then
           return One( SqrtField );
        elif k=2 then
           return -One(SqrtField);
        elif k=3 then
           return z^40;
        elif k=4 then
           return z^30;
        elif k=5 then
           return z^24;
        elif k=6 then
           return z^20;
        elif k=8 then
	       return z^15;
	    elif k=9 then
	       return Sqroot( 3, PRU(3) );
        elif k=10 then
           return z^12;
        elif k=12 then
	       return z^10;
        elif k=15 then
           return z^8;
        elif k=20 then
           return z^6;
        elif k=24 then
           return z^5;
        elif k=40 then
           return z^3;
        elif k=60 then
           return z^2;
        elif k=120 then
           return z;
        fi;

end;


InstallMethod( \*,
   "for rational and SqrtField elements",
   [IsCyclotomic, IsSqrtFieldElement], 
function(x,y)
local b,i,ex,m,z,u,v;

      if IsGaussRat(x) then
         if x = 0 then return Zero(SqrtField); fi;
         if x = 1 then return y; fi;
         b := List(y![1],ShallowCopy);
         for i in b do i[1] := x*i[1]; od;
         return SqrtField_objectify( b );          
      fi;
	 
      if IsOne(y) then

         ex:= ExtRepOfObj(x);
         m:= Length(ex);
         if not m in [1,2,3,4,5,6,8,9,10,12,15,20,24,40,60,120] then
            Error("We cannot currently deal with the cyclotomic RPU( ",m,")");
         fi;
         z:= PRU(m);
     	 u:= Zero( SqrtField );
         v:= One( SqrtField );
	     for i in [0..m-1] do
	         u:= u + (ex[i+1]*One(SqrtField))*v;
	         v:= v*z;
	     od;
	     return u;
      fi;

      return (x*One(SqrtField))*y;

end);


#########################################################################

InstallMethod( \*,
   "for SqrtField and rational elements",
   [IsSqrtFieldElement, IsCyclotomic], 
function(x,y)
   return y*x;      
end);

#########################################################################

InstallMethod( ZeroOp,
   "for SqrtField element",
   [IsSqrtFieldElement], 
function(x)
   return Zero(SqrtField);     
end);

#########################################################################

InstallMethod( OneOp,
   "for SqrtField element",
   [IsSqrtFieldElement], 
function(x)
   return One(SqrtField);   
end);

#########################################################################

InstallMethod( \+,
   "for SqrtField and SqrtField elements",
   [IsSqrtFieldElement, IsSqrtFieldElement], 10000,
function(x,y)
local a,b,i,j,e,len,pos;

   if x = Zero(SqrtField) then return y; fi;
   if y = Zero(SqrtField) then return x; fi;
   if x![2] then
      if y![2] then
         return SqrtField_objectify([[x![1][1][1]+y![1][1][1], []]]);
      else 
         return x![1][1][1] + y;
      fi;
   fi; 
   if y![2] then return y![1][1][1] + x; fi;

   if Length(x![1])<=Length(y![1]) then
      a := List(x![1],ShallowCopy);
      b := List(y![1],ShallowCopy);
   else
      a := List(y![1],ShallowCopy);
      b := List(x![1],ShallowCopy);
   fi;
   i := 1;
   for j in [1..Length(a)] do
       pos:= PositionSorted( b, a[j],
                  function( m1, m2 ) return m1[2]<m2[2]; end );
       if pos <= Length(b) and b[pos][2] = a[j][2] then
          b[pos][1]:= b[pos][1]+a[j][1];
	  if IsZero( b[pos][1] ) then
	     RemoveElmList( b, pos );
	  fi;
       else
          InsertElmList( b, pos, a[j] );
       fi;
   od;
   
   if b = [] then return Zero(SqrtField); fi;
   return SqrtField_objectify( b );          
end);

#########################################################################

InstallMethod( \+,
   "for Gaussian rational and SqrtField elements",
   [IsCyclotomic, IsSqrtFieldElement], 
function(x,y)
local t;
   if not IsGaussRat(x) then
      Error("cyclotomic has to be Gaussian rational");
   fi;
   if x=0 then return y; fi;
   t := List(y![1],ShallowCopy);
   if t[1][2] = [] then
      t[1][1] := t[1][1] + x;
      if t[1][1] = 0 then t := t{[2..Length(t)]}; fi;
      if t = [] then return Zero(SqrtField); fi;
   else
     #add [y,[]] at position 1
      CopyListEntries( t, 1, 1, t, 2, 1, Length(t) );
      t[1] := [x,[]];
   fi;
   return SqrtField_objectify( t );
end);

#########################################################################

InstallMethod( \+,
   "for SqrtField and rational elements",
   [IsSqrtFieldElement, IsCyclotomic], 
function(x,y)
   return y+x;
end);

#########################################################################

InstallMethod( \AINV,
   "for SqrtField elements",
   [IsSqrtFieldElement], 
function(x)
local a, i;
   a := List(x![1],ShallowCopy);
   for i in a do i[1] := -i[1]; od;
   return SqrtField_objectify( a );   
end);

#########################################################################

InstallMethod( \=,
   "for SqrtField elements",
   [IsSqrtFieldElement, IsSqrtFieldElement], 
function(x,y)
   return x![1] = y![1];
end);
#########################################################################

InstallMethod( \=,
   "for SqrtField and rational elements",
   [IsSqrtFieldElement, IsRat], 
function(x,y)
   if not x![2] then return false; fi;
   return SqrtFieldMakeRational(x) = y;
end);

#########################################################################

InstallMethod( \=,
   "for rational and SqrtField element",
   [IsRat, IsSqrtFieldElement], 
function(x,y)
   if not y![2] then return false; fi;
   return SqrtFieldMakeRational(y) = x;
end);

#########################################################################

collect:= function( mons, coefs )

        # mons: monomials, coefs: coefficients
	# cfs: coefficients
	# we perform collection, the result is a SqrtFieldElt

      local mns, cfs, mns0, cfs0, mn, c, pos, i, pos0, p, q, exp, tl,
            mid, st, res, neg;

      mns:= ShallowCopy( mons );
      cfs:= ShallowCopy( coefs ); 

      mns0:= [ ];
      cfs0:= [ ];
      while Length(mns) > 0 do
#Print("mns  ",mns,"\n",cfs,"\n");
#Print("mns0  ",mns0,"\n",cfs0,"\n");      
          mn:= mns[1];
	  c:= cfs[1];
	  RemoveElmList( mns, 1 );
	  RemoveElmList( cfs, 1 );
	  pos:= [ ]; neg:= [ ];
	  for i in mn do
	      if IsList( i ) then
	         Add( neg, i );
              else
	         Add( pos, i );
              fi;
	  od;
	  SORT_LIST(pos);
	  SORT_LIST(neg); neg:= Reversed( neg );
	  pos0:= [ ];
	  i:= 1;
	  while i <= Length(pos) do
	      if i < Length(pos) and pos[i] = pos[i+1] then
	         c:= c*pos[i];
		 i:= i+2;
              else
	         Add( pos0, pos[i] );
		 i:= i+1;
              fi;
	  od;

	  q:= PositionProperty([1..Length(neg)-1],i -> neg[i][1]=neg[i+1][1] );
	  if q <> fail then
	     p:= SqrtFieldFam!.newp[ -neg[q][1] ];
	     exp:= neg[q][2]+neg[q+1][2];
	     if exp < p then
	        Add( cfs, c );
		tl:= neg{[1..q-1]};
		Add( tl, [ neg[q][1], exp ] );
		Append( tl, neg{[q+2..Length(neg)]} );
		Add( mns, Concatenation( pos0, tl ) );
             elif exp = p then
	        mid:= SqrtFieldFam!.newsq[ -neg[q][1] ];
		if IsRat(mid) then
		   Add( cfs, c*mid );
		   tl:= neg{[1..q-1]};
		   Append( tl, neg{[q+2..Length(neg)]} );
		   Add( mns, Concatenation( pos0, tl ) );
		else   
    	           st:= neg{[1..q-1]};
	           tl:= neg{[q+2..Length(neg)]};
	           for i in mid do
	               Add( cfs, c*i[1] );
		       Add( mns, Concatenation( pos0, st, i[2], tl ) );
                   od;
		fi;   
	     else # that is, exp>p
	        mid:= SqrtFieldFam!.newsq[ -neg[q][1] ];
		if IsRat(mid) then
		   Add( cfs, c*mid );
		   tl:= neg{[1..q-1]};
		   Add( tl, [ neg[q][1], exp-p ] );
		   Append( tl, neg{[q+2..Length(neg)]} );
		   Add( mns, Concatenation( pos0, tl ) );
		else   
    	           st:= neg{[1..q-1]};
		   Add( st, [ neg[q][1], exp-p ] );
	           tl:= neg{[q+2..Length(neg)]};
	           for i in mid do
	               Add( cfs, c*i[1] );
		       Add( mns, Concatenation( pos0, st, i[2], tl ) );
                   od;
		fi;
             fi;
	  else
	     Add( cfs0, c );
	     Add( mns0, Concatenation( pos0, neg ) );
	  fi;
      od;
#Print("mns0  ",mns0,"\n",cfs0,"\n");      
      SORT_PARA_LIST( mns0, cfs0 );

      res:= [ ];
      i:= 1; 
      while i <= Length(mns0) do
          mn:= mns0[i]; c:= cfs0[i];
	  i:= i+1;
	  while i <= Length(mns0) and mns0[i] = mn do
	        c:= c+cfs0[i];
	        i:= i+1;
          od;
	  if not IsZero(c) then 
    	     Add( res, [ c, mn ] );
	  fi;
      od;
      if res = [] then return Zero(SqrtField); fi;
      return SqrtField_objectify( res );

end;


InstallMethod( \*,
   "for SqrtField and SqrtField elements",
   [IsSqrtFieldElement, IsSqrtFieldElement],
   function(x,y)
   local a, b, i, j, mns, cfs;

      if x=Zero(SqrtField) or y=Zero(SqrtField) then 
         return Zero(SqrtField); 
      fi;
      if x=One(SqrtField) then return y; fi;
      if y=One(SqrtField) then return x; fi;

      a   := x![1];
      b   := y![1];

      if x![2] and y![2] then
         return SqrtField_objectify([[a[1][1]*b[1][1],[]]]);
      fi; 

      mns := [ ];
      cfs := [ ];

      for i in a do
         for j in b do
             Add( mns, Concatenation( i[2], j[2] ) );
	     Add( cfs, i[1]*j[1] );
         od;
      od;

      # now we start collecting...
      return collect( mns, cfs );
end);


#######################################################################

minpol:= function( el )

      local bas, cfs, xk, ex, cf, added, i, k, p, u;

      bas:= [ [] ];
      cfs:= [ [1] ];
      xk:= One(SqrtField);
      while true do
      #Print(xk,"\n",el,"\n\n");
         xk:= xk*el;
	 ex:= xk![1];
	 cf:= List( bas, x -> 0 );
	 added:= false;
	 for i in [1..Length(ex)] do
	     p:= PositionSorted( bas, ex[i][2] );
	     if p<= Length(bas) and bas[p] = ex[i][2] then
	        cf[p]:= ex[i][1];
             else
	        InsertElmList( bas, p, ex[i][2] );
		for k in [1..Length(cfs)] do
		    InsertElmList( cfs[k], p, 0 );
		od;
		InsertElmList( cf, p, ex[i][1] );
		added:= true;
             fi;
	 od;
	 Add( cfs, cf );
	 if not added then # see whether we find an elm in kernel
	    if RankMat(cfs) < Length(cfs) then 
               u:= NullspaceMat(cfs);	    
               if Length(u) > 0 then
                  u:= u[1];
	          u:= u/u[ Length(u) ];
	          return u;
	       fi;
	    fi;
	 fi;
      od;
end;      

##########################################################################

InstallMethod( InverseOp,
   "for SqrtField elements",
   [IsSqrtFieldElement], 
function(el)
   local c, n, i, inv;
   
   c:= minpol(el);
   n:= Length(c);
   inv:= c[n];
   for i in [n-1,n-2..2] do
       inv:= c[i]+el*inv;
   od;
   inv:= -inv/c[1];
   return inv*One(SqrtField);       
end);


#########################################################################

InstallMethod( \AINV_MUT,
   "for SqrtField elements",
   [IsSqrtFieldElement],
   function(x)
   local a, i;
      a := List(x![1],ShallowCopy);
      for i in a do i[1] := -i[1]; od;
      return SqrtField_objectify( a );
end);

#########################################################################

InstallMethod( \<,
   "for SqrtField elements",
   [IsSqrtFieldElement, IsSqrtFieldElement],
   function(x,y)
      return x![1] < y![1];
end);

#########################################################################

InstallMethod( \<,
   "for SqrtField element and Rational",
   [IsSqrtFieldElement, IsRat],
   function(x,y)
      if not x![2] then TryNextMethod(); fi;
      return SqrtFieldMakeRational(x) < y;
end);

########################################################################

InstallMethod( \<,
   "for rational and SqrtField element",
   [IsRat, IsSqrtFieldElement],
   function(x,y)
      if not y![2] then TryNextMethod(); fi;
      return SqrtFieldMakeRational(y) > x;
end);


#########################################################################


InstallGlobalFunction(SqrtFieldMinimalPolynomial, function(el)
#returns polynomial f of min deg over GaussianRats such that f(el)=0; 
local b, i, j, bas, mat, elts, a, cfs, l,cf, pos, ns, inv, res, v, nrp, 
      want, pol, x;
   if not IsSqrtFieldElement(el) then
      Error("input has to be SqrtField elt");
   fi;
   if el=Zero(SqrtField) then Error("elt must be nonzero"); fi;
   a    := List(el![1],ShallowCopy);
   a    := SqrtField_objectify( a );
   i    := 1;
   elts := [One(SqrtField)];
   bas  := [[]];
   cfs  := [ [1] ];
   l    := Length(bas);
   nrp  := Length(Collected(Flat(List(a![1],x->x[2]))));
   want := 2^Minimum([nrp,Length(Filtered(a![1],x->not x[2]=[]))]);
   #if want > 16 then
   #   Info(InfoSqrtField,1,"need ",want," powers for computing the inverse");
   #fi;
   repeat
      i       := i+1;
      elts[i] := elts[i-1]*a; 
      cfs[i]  := ListWithIdenticalEntries(l,0);
      for j in elts[i]![1] do
         pos := Position(bas,j[2]);
         if pos=fail then 
            Add(bas,j[2]); 
            l   := l+1;  
            pos := l; 
            for v in cfs do v[l] := 0; od;
         fi;
         cfs[i][pos] := j[1];
      od;
      #if i mod 10 = 0 then
      #   Info(InfoSqrtField,2,"  - computed ",i," elements");
      #fi;
   until i-1 = want;
   ns   := NullspaceMat(cfs);
   res  := ns[1]*One(SqrtField);
   x    := Indeterminate(SqrtField);
   elts := [One(SqrtField)];
   Append(elts, List([2..Length(res)],i->x^(i-1)));
   pol  := Sum(List([1..Length(res)],i->res[i]*elts[i]));
   return pol;       
end);

#########################################################################


InstallGlobalFunction(SqrtFieldEltByRationalSqrt, function(e)
local sgn;
   if not IsRat(e^2) then Error("input has to be Sqroot of a rational"); fi;
   if not e-Sqrt(e^2) = 0 then sgn := -1; else sgn := 1; fi;
   return sgn*Sqroot(e^2);
end); 

##########################################################################


InstallGlobalFunction(SqrtFieldEltRealAndComplexPart, function(e)
local cc;
      cc:= ComplexConjugate(e);
      return [ (e+cc)/2, (e-cc)/(2*E(4)) ];
end); 
##########################################################################


InstallGlobalFunction(IsPosSqrtFieldElt, function(e)
local real, imag, i, makecf, cf, r;
   if not IsSqrtFieldElement(e) or not Length(e![1]) = 1 or
      not ForAll(e![1],i->IsRat(i[1])) then 
      Error("input must be a real SqrtFieldElement with one monom"); 
   fi;
   return e![1][1][1]>0;
end); 

##########################################################################


InstallGlobalFunction(CoefficientsOfSqrtFieldElt, function(e)
local cfs, i, x;
   if not IsSqrtFieldElement(e) then 
      Error("input must be SqrtFieldElement"); 
   fi;
   x := List(e![1],ShallowCopy);
   return List(x,i->[i[1],Product(i[2])]);
end); 

##########################################################################


InstallGlobalFunction(SqrtFieldEltByCoefficients, function(e)
local elt, i;
   elt := Zero(SqrtField);
   for i in e do
      if not IsGaussRat(i[1]) and IsRat(i[2]) then 
          Error("wrong format of coefficients vector"); 
      fi;
      elt := elt + i[1]*Sqroot(i[2]);
   od;
   return elt;
end); 

##########################################################################


InstallGlobalFunction(SqrtFieldEltToCyclotomic, function(e)
local makemon;
   if not IsSqrtFieldElement(e) then
      Error("input has to lie in SqrtField");
   fi;
   if Length(e![1])=1 and e![1][1][2]=[] then return e![1][1][1]; fi;
  #Info(InfoSqrtField,1,"Warning: might not work for large cyclotomics.");
   makemon := function(mon)
      if mon[2]=[] then return mon[1]; fi;
      return mon[1]*Sqrt(Product(mon[2]));
   end;
   return Sum(List(e![1],makemon));
end); 

##########################################################################

InstallMethod( DefaultFieldOfMatrix,
    "method for a matrix over SqrtField",
    [ IsMatrix and IsSqrtFieldElementCollColl ],
function( mat )
   return SqrtField;
end );

##########################################################################

InstallMethod( DefaultFieldByGenerators,
    "method for a list over SqrtField",
    [ IsList and IsSqrtFieldElementCollection ],
function( mat )
   return SqrtField;
end );

##########################################################################

comp_conj:= function( elt, bool )

      # elt in extrep, bool is true if elt is of the form [ -i, k ]
      # we return the complex conjugate in wrapped form

      local i, k, b, a, res, j, mon, pos, m0, c, ap;

      if bool then
         i:= -elt[1]; k:= elt[2];
         if IsBound( SqrtFieldFam!.newconj[i] ) then
	    b:= SqrtFieldFam!.newconj[i];
	    return b^k;
	 else
	    a:= List( SqrtFieldFam!.newsq[i], ShallowCopy );
            if ForAll( a, x -> PositionProperty( x[2], IsList )=fail ) then
	       for j in [1..Length(a)] do
	           a[j][1]:= ComplexConjugate( a[j][1] );
	       od;
	       b:= SqrtField_objectify( a );
#Print("b1:...  ",b,"\n\n");	       	       
               b:= nRoot( SqrtFieldFam!.newp[i], b );
	    else
	       b:= comp_conj( a, false );
#Print("b2:...  ",b,"\n\n");	       
	       b:= nRoot( SqrtFieldFam!.newp[i], b );
            fi;
	    SqrtFieldFam!.newconj[i]:= b;
	    c:= b^k;
#Print("start approx...\n");
#Print(elt,"\n");
#Print( SqrtField_objectify([ [1, [elt]] ]), "\n" );
	    ap:= approx( SqrtField_objectify([ [1, [elt]] ]) );
#Print(ap,"\n"); 
	    ap[2]:= -ap[2];
	    ap:= ap-approx(c);
#Print("done approx...  ",ap,"\n");	    	    
	    if ap[1]^2+ap[2]^2 > eps_0 then
	       if SqrtFieldFam!.newp[i] = 2 then
	       	  SqrtFieldFam!.newconj[i]:= -b;
	          return -c;
	       else
	          return c;
	          Print("ERROR in Complex Conjugation!!!\n");
	       fi;
	    else
	       return c;
	    fi;
	 fi;
      fi;

      res:= Zero( SqrtField );
      for i in [1..Length(elt)] do
          mon:= elt[i][2];
	  pos:= PositionProperty( mon, IsList );
	  if pos = fail then
	     res:= res+SqrtField_objectify( [ [ComplexConjugate(elt[i][1]),
	                   mon] ] );
	  else
	     m0:= mon{[1..pos-1]};
	     a:= SqrtField_objectify( [ [ComplexConjugate(elt[i][1]), m0] ] );
	     for j in [pos..Length(mon)] do
	         a:= a*comp_conj( mon[j], true );
             od;
	     res:= res+a;
	  fi;
      od;
      return res;

end;

      
	       
	      


InstallMethod( ComplexConjugate,
    "method for SqrtField elt",
    [ IsSqrtFieldElement ],
function( el )
return comp_conj( el![1], false );

end );

##########################################################################

InstallMethod( Sqrt,
    "method for SqrtField elt",
    [ IsSqrtFieldElement ],
function( el )
   return  Sqroot(el);
end );

##########################################################################

InstallMethodWithRandomSource( Random,
    "for a random source and an SqrtField ",
    [ IsRandomSource, IsSqrtField ],
function( rs, F )
   return Sum(List([1..15],x-> Random(rs,Rationals)*Sqroot(Random(rs,Rationals))));
   
end );

##########################################################################

InstallMethod( AbsoluteValue,
    "for SqrtField element",
    [ IsSqrtFieldElement ], 10000,
function( el )
local x,i;
   x := List(el![1],ShallowCopy);
   if not (Length(x)=1 and IsRat(x[1][1])) then
      Error("elt must be a real SqrtFieldElt with one summand");
   fi;
   for i in x do i[1] := AbsoluteValue(i[1]); od;
   return SqrtField_objectify( x );
end );

##########################################################################

InstallMethod( String,
    "for SqrtField element",
    [ IsSqrtFieldElement ],
function( el )
local elt, l, i, printmon, printcf, str;
   elt := List(el![1],ShallowCopy);
   str := "";
   l   := Length(elt);
   printmon := function(mon)
      if mon[1] = 1 then 
         Append(str, "Sqroot("); 
         Append(str,String(Product(mon[2])));
         Append(str,")");
      elif mon[1]<0 or not (IsRat(mon[1]) or IsRat(E(4)*mon[1])) 
                    or (IsRat(E(4)*mon[1]) and E(4)*mon[1]>0) then
         Append(str,"(");
         Append(str,String(mon[1]));
         Append(str,")*Sqroot(");
         Append(str,String(Product(mon[2])));
         Append(str,")");
      else 
         Append(str, String(mon[1]));
         Append(str,"*Sqroot(");
         Append(str,String(Product(mon[2])));
         Append(str,")");
      fi;
   end;
   if l=1 then
      if elt[1][2]=[] then Append(str,String(elt[1][1])); else printmon(elt[1]); fi;
   else
      if elt[1][2]=[] then 
         Append(str,String(elt[1][1]));
         Append(str," + "); 
      else 
         printmon(elt[1]); 
         Append(str," + ");
      fi;
   fi;
   for i in [2..l-1] do printmon(elt[i]); Append(str," + "); od;
   if l>1 then printmon(elt[l]); fi;  
   return str;
end );

##########################################################################

InstallMethod( String,
    "for SqrtField",
    [ IsSqrtField ],
function( el )
   return "SqrtField";
end );

##########################################################################


InstallGlobalFunction(SqrtFieldEltByCyclotomic, function(arg)
local e, cj, re, im, solvereal, res, r1, r2, 
      sqfree, whichSqfreeRootsInCF;
   e := arg[1];
   if e in SqrtField then return e; fi;
   if not IsCyclotomic(e) then 
      Error("input must be Cyclotomic"); 
   fi;
   if IsGaussRat(e) then return e*One(SqrtField); fi;

   whichSqfreeRootsInCF := function(n)
  #returns also one for technical reasons
   local sqfree, eeven, res;
      sqfree := x -> ForAll(Collected(FactorsInt(x)),i->i[2]=1);
      eeven  := x -> Length( Filtered(List(Collected(FactorsInt(x)), i->i[1]),
                             p -> p mod 4 = 3)) mod 2 = 0;
      res    := Filtered(DivisorsInt(n),x->sqfree(x));
      if IsOddInt(n) or (IsInt(n/2) and IsOddInt(n/2)) then
         return Filtered(res,x->IsOddInt(x) and eeven(x));
      fi;
      if n mod 4 = 0 and not n mod 8 =0 then
         return Filtered(res,IsOddInt);
      fi;
      return res;
   end;

  #check if input is sqrt of rational
  #test is expensive, e.g. for e=Sqrt(1231)+Sqrt(17);;
  #if IsRat(e^2) then return SqrtFieldEltByRationalSqrt(e); fi;

  #split into real and complex part
   cj    := ComplexConjugate(e);;
   re    := 1/2*(e+cj);;
   im    := 1/2*E(4)^3*(e-cj);;

  #the main function which requires a real input   
   solvereal := function( r )
   local n, pr, prsq, bas, mat, ns;  
      if IsRat(r) then return r*One(SqrtField); fi;
      n     := Conductor(r);
      pr    := whichSqfreeRootsInCF(n);
      prsq  := List(pr,Sqrt);
      bas   := Basis(CF(n));
      mat   := List(prsq,x->Coefficients(bas,x));
      ns    := SolutionMat(mat,Coefficients(bas,r));
      if ns = fail then return fail; fi;
      return Sum(List([1..Length(pr)],i->ns[i]*Sqroot(pr[i])));
   end;

  #try to solve real and complex part
   r1  := solvereal(re); if r1 = fail then return fail; fi;
   r2  := solvereal(im); if r2 = fail then return fail; fi;
   res := r1 + E(4)*r2;
 ###TEST
  #Info(InfoSqrtField,1,"test res");
  #if not e = SqrtFieldEltToCyclotomic(res) then
  #   Error("res wrong");
  #fi;
   return res;
end); 


#########################################################################


InstallGlobalFunction( SqrtFieldPolynomialToRationalPolynomial, function(e)
local x,i,j,cf;
   x  := Indeterminate(Rationals);
   cf := List(CoefficientsOfUnivariatePolynomial(e),SqrtFieldEltToCyclotomic);
   return Sum(List([1..Length(cf)],i-> cf[i]*x^(i-1)));
end); 


#########################################################################


InstallGlobalFunction( SqrtFieldRationalPolynomialToSqrtFieldPolynomial, function(e)
local x,i,j,cf;
   x  := Indeterminate(SqrtField);
   cf := List(CoefficientsOfUnivariatePolynomial(e),SqrtFieldEltByCyclotomic);
   return Sum(List([1..Length(cf)],i-> cf[i]*x^(i-1)));
end); 



##########################################################################

InstallOtherMethod( Factors,
    "for SqrtField",
    [ IsPolynomialRing, IsUnivariatePolynomial ],
function( R, f )
local rf;
if LeftActingDomain(R) = SqrtField and 
      IsSqrtFieldElementCollection(CoefficientsOfUnivariatePolynomial(f)) then
      rf := SqrtFieldPolynomialToRationalPolynomial(f);
      rf := Factors(rf);
      return List(rf,SqrtFieldRationalPolynomialToSqrtFieldPolynomial);
   else
      TryNextMethod();
   fi;
end );


